import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext
from xrnn.layers import Dense
from xrnn.losses import MSE
from xrnn import optimizers
from xrnn import ops
import pytest


@pytest.fixture
def layer():
    ops.random.seed(0)
    layer = Dense(16)
    return layer


def two_iteration_train_step(layer, opt):
    loss = MSE()
    labels = ops.ones((32, 16))
    features = ops.ones((32, 1))
    for _ in range(2):
        layer.backward(loss.backward(labels, layer(features)))
        opt.update_params(layer)
        opt.iterations += 1


def step_decay(_: float, clr: float, itr: int, __: int) -> float:
    if (itr + 1) % 10 == 0:
        return clr / 2
    return clr


class TestOptimizer:

    def test_initialize_layer_state(self):
        layer = Dense(16, 100)
        layer.build()
        opt = optimizers.Optimizer()
        opt.initialize_layer_state(layer, momentums=False)
        assert layer.weights_cache.shape == layer.weights.shape
        assert not hasattr(layer, 'weight_momentums')
        prev_id = id(layer.weights_cache)
        opt.initialize_layer_state(layer)
        assert id(layer.weights_cache) == prev_id
        new_opt = optimizers.Optimizer()
        new_opt.initialize_layer_state(layer)
        assert layer.weight_momentums.shape == layer.weights.shape
        assert id(layer.weights_cache) == prev_id

    @pytest.mark.parametrize(
        "decay_func, raises",
        [
            (lambda x: x / 10, pytest.raises(RuntimeError)),
            (lambda ilr, clr, itr, epoch: 1 / 0, pytest.raises(RuntimeError)),
            (lambda ilr, clr, itr, epoch: '1', pytest.raises(TypeError)),
            (lambda irl, clr, itr, epoch: clr * epoch, nullcontext())
        ]
    )
    def test_validate_decay_function(self, decay_func, raises):
        with raises:
            optimizers.Optimizer.validate_decay_function(decay_func)

    @pytest.mark.parametrize(
        "lr, decay, decay_func, expected",
        [
            (0.1, 0, None, 0.1),
            (0.1, 0.1, None, 0.009174),
            (0.1, 0, lambda lr, clr, itr, epoch: lr / (itr + 1), 0.001),
            (0.1, 0, step_decay, 9.8e-5),
        ]
    )
    def test_update_learning_rate(self, lr, decay, decay_func, expected):
        opt = optimizers.Optimizer(lr, decay, decay_func)
        for _ in range(100):
            opt.update_learning_rate()
            opt.iterations += 1
        assert round(opt.current_lr, 6) == expected

    @pytest.mark.parametrize(
        "opt_type, kwargs, n_elements",
        [
            ('SGD', {'momentum': 0.9}, 7),
            ('Adagrad', {'decay': 0.3}, 6),
            ('RMSprop', {'rho': 0.5}, 7),
            ('Adam', {'beta_2': 0.5}, 8),
        ]
    )
    def test_get_config(self, opt_type, kwargs, n_elements):
        opt_config = getattr(optimizers, opt_type)(0.2, **kwargs).get_config()
        assert len(opt_config) == n_elements
        assert opt_config['learning_rate'] == 0.2
        key, value = kwargs.popitem()
        assert opt_config[key] == value
        assert opt_config['type'] == opt_type

    @pytest.mark.parametrize(
        'optimizer_type, kwargs, error',
        [
            ('Adam', {'beta_1': 0.2}, nullcontext()),
            ('SGD', {'momentum': 0.9}, nullcontext()),
            ('RMSprop', {'rho': 0.}, nullcontext()),
            ('Adagrad', {'learning_rate': 0.2}, nullcontext()),
            ('Adam', {'Sheesh': 0.}, pytest.raises(TypeError)),
        ]
    )
    def test_from_config(self, optimizer_type, kwargs, error):
        opt_config = {
            'learning_rate': 0.1,
            'current_lr': 0.01,
            'decay': 0.9,
            'iterations': 100,
            'epoch': 2}
        opt_config.update(kwargs)
        opt_config.update({'type': optimizer_type})
        with error:
            opt = getattr(optimizers, optimizer_type).from_config(opt_config)
            assert opt.learning_rate == opt_config['learning_rate']
            assert opt.current_lr == 0.01
            assert opt.decay == 0.9
            assert opt.iterations == 100
            assert opt.epoch == 2
            key, val = kwargs.popitem()
            assert getattr(opt, key) == val
        with pytest.raises(TypeError):
            optimizers.Optimizer.from_config(opt_config)


# The numbers that are checked against have been obtained after two optimization steps using Tensorflow optimizers
# constructed with the same parameters as this package's optimizers.
# The reason why two optimization steps are performed is that for optimizers that use momentum or cache, the first
# step is a warm-up step/initialization step, and the second one is when they actually update the weights using them.
# The ideal way to test optimizers is to check their results against Tensorflow optimizers during testing, but having
# Tensorflow as a dependency for running the tests is hugely inefficient.


@pytest.mark.parametrize(
    "opt, expected",
    [
        (optimizers.SGD(), 0.011385817),
        (optimizers.SGD(momentum=0.9), 0.012500793),
        (optimizers.Adagrad(), 0.010617446),
        (optimizers.RMSprop(), 0.014359113),
        (optimizers.Adam(), 0.010911059)
    ]
)
def test_update_params(opt, expected, layer):
    two_iteration_train_step(layer, opt)
    assert layer.weights.mean() == pytest.approx(expected, 1e-5, 3e-7)



import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext
import importlib
import platform
import unittest.mock
import ctypes

import pytest

from xrnn import ops
from xrnn import c_layers

# Import c_layers and reload it for every test, because imported modules are cached, and importing them again will just
# bind the already imported module (c_layers) to this module's namespace. That's a problem because the module level code
# has been already executed, meaning that tests for branching code aren't going to get executed.

CURR_OS = platform.system()


def _test_os(mocked_call, os, err):
    context_manager = pytest.raises(err) if err else nullcontext()
    with context_manager:
        mocked_call.return_value = os
        importlib.reload(c_layers)


@unittest.mock.patch('platform.system')
def test_os_lib_extension(mocked_system_call):
    if CURR_OS == 'Windows':
        _test_os(mocked_system_call, 'Linux', FileNotFoundError)
    else:
        _test_os(mocked_system_call, 'Windows', FileNotFoundError)
    _test_os(mocked_system_call, 'CustomOS', OSError)


def test_c_functions_type_checking():
    importlib.reload(c_layers)
    args = (ops.ones((32, 28, 28, 3)), ops.ones((16, 3, 3, 3)), ops.ones((16,)), ops.ones((32, 26, 26, 16)),
            3, 3, 1, 1, 32, 26, 26, 16, 28, 28, 3, True)
    # Test too many arguments.
    try:
        # We wrap this with a try block because Linux might or might not raise an exception (RuntimeError), if it does,
        # we catch that and check if it's RuntimeError (Linux/macOS) or TypeError (windows).
        c_layers.convForwardF(*args, 1)
    except (RuntimeError, TypeError):
        pass
    except Exception as e:
        assert False, f'Got an unexpected exception: {e}'
    # Test not enough arguments.
    with pytest.raises(TypeError):
        c_layers.convForwardF(*args[:-1])
    # Test wrong array dtype.
    with pytest.raises(ctypes.ArgumentError):
        c_layers.convForwardF(ops.ones(args[0].shape, 'f8'), *args[1:])
    # Test wrong number type
    with pytest.raises(ctypes.ArgumentError):
        c_layers.convForwardF(*args[:-2], 3.1, True)
    c_layers.convForwardF(*args)



import pytest

from xrnn import config
from xrnn import layer_utils
from xrnn import layers

nhwc_shape = (128, 32, 32, 3)
nchw_shape = (128, 3, 32, 32)
dense_shape = (128, 256)  # Batch size 128, features/dimensions 256
image_batch_nbytes = 128 * 32 * 32 * 3 * 4
dense_batch_nbytes = 128 * 256 * 4


@pytest.mark.parametrize(
    "input_shape, window_size, strides, expected, image_format",
    [
        (nhwc_shape, (5, 5), (1, 1), (32, 32), 'channels_last'),
        (nhwc_shape, (7, 7), (2, 2), (16, 16), 'channels_last'),
        (nchw_shape, (3, 9), (3, 4), (11, 8), 'channels_first'),
        (nchw_shape, (5, 9), (3, 1), (11, 32), 'channels-first'),
        (nhwc_shape, (3, 3), (2, 2), (16, 16), 'channels-last'),
        (nhwc_shape, (11, 11), (4, 4), (8, 8), 'channels_last'),
        (nhwc_shape, (4, 4), (6, 6), (6, 6), 'channels_last'),
        (nchw_shape, (4, 4), (6, 6), (6, 6), 'channels_first'),
    ]
)
def test_padded_output_shape(input_shape, window_size, strides, expected, image_format):
    """This function tests both `layer_utils.compute_spatial_output_shape` and `layer_utils.calculate_padding_on_sides`
    at the same time because they depend on each other, and we are more interested in the output shape rather than the
    return of each function individually."""
    config.set_image_data_format(image_format)
    assert layer_utils.compute_spatial_output_shape(
        input_shape,
        window_size,
        strides,
        layer_utils.calculate_padding_on_sides(input_shape, window_size, strides)
    ) == expected


@pytest.mark.parametrize(
    "layer, expected_params, expected_activations",
    [
        (layers.Dense(8), 8 * 4 + 8 * 256 * 4, dense_batch_nbytes),
        (layers.Dropout(0.5), 0, image_batch_nbytes * 2),
        (layers.Conv2D(8, 3, 2), 896, image_batch_nbytes),
        (layers.Conv2D(8, 3, 2, 'same'), 896, 128 * 33 * 33 * 3 * 4),
        (layers.Conv2D(8, 5, 3, 'same'), 5 * 5 * 8 * 3 * 4 + 8 * 4, 128 * 35 * 35 * 3 * 4),
        (layers.MaxPooling2D(2, 1), 0, image_batch_nbytes),
        (layers.MaxPooling2D(2, 1, 'same'), 0, 128 * 33 * 33 * 3 * 4),
        (layers.AvgPooling2D(2, 2), 0, image_batch_nbytes),
        (layers.AvgPooling2D(2, 2, 'same'), 0, 128 * 32 * 32 * 3 * 4),
        (layers.BatchNormalization(), 48, image_batch_nbytes * 2),
        (layers.Flatten(), 0, image_batch_nbytes)
    ]
)
def test_layer_memory_consumption(layer, expected_params, expected_activations, request):
    # `request` is a `pytest` built-in function that is passed to tests as an argument.
    for input_shape, image_format, dtype in zip(
            (nhwc_shape, nchw_shape), ('channels_last', 'channels_first'), ('f', 'd')):
        config.set_image_data_format(image_format)
        config.set_default_dtype(dtype)
        if isinstance(layer, layers.BatchNormalization):
            grads_mem = nhwc_shape[-1] * 4 * 2
            expected_total = expected_activations + expected_params + grads_mem * 3
        else:
            grads_mem = expected_params
            expected_total = expected_activations + expected_params * 4
        if dtype == 'd':
            expected_params *= 2
            grads_mem *= 2
            expected_total *= 2
        input_shape = dense_shape if isinstance(layer, layers.Dense) else input_shape
        if getattr(layer, 'built') is False:
            layer.build(input_shape)
        params_mem_consumption, *_, total_mem_consumption = layer_utils.layer_memory_consumption(
            request.getfixturevalue(layer) if isinstance(layer, str) else layer, input_shape, True, True)
        assert params_mem_consumption == expected_params
        assert total_mem_consumption == expected_total
        params_mem_consumption, *_, total_mem_consumption = layer_utils.layer_memory_consumption(
            request.getfixturevalue(layer) if isinstance(layer, str) else layer, input_shape, False, True)
        assert params_mem_consumption == total_mem_consumption == expected_params
        params_mem_consumption, *_, total_mem_consumption = layer_utils.layer_memory_consumption(
            request.getfixturevalue(layer) if isinstance(layer, str) else layer, input_shape, True, False)
        assert params_mem_consumption == expected_params
        assert total_mem_consumption == expected_total - grads_mem


def test_make_unique_name():
    class Test:
        pass

    assert layer_utils.make_unique_name(1) == 'int_0'
    assert layer_utils.make_unique_name(1) == 'int_1'
    assert layer_utils.make_unique_name(Test()) == 'Test_0'
    assert layer_utils.make_unique_name(Test()) == 'Test_1'
    with pytest.raises(TypeError):
        layer_utils.make_unique_name(object)


@pytest.mark.parametrize(
    "time, expected",
    [
        (0.2, '20 ms'),
        (32.33, '32 sec'),
        (32.53, '33 sec'),
        (121.32, '2.0 min'),
        (1.52 * 60 * 60, '1.5 hrs'),
        (34.44 * 60 * 60 * 24, '34.44 days')
    ]
)
def test_time_unit_converter(time, expected):
    assert layer_utils.time_unit_converter(time) == expected



import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext

from xrnn import config
from xrnn import ops
import numpy as np
import pytest


@pytest.mark.parametrize(
    "dtype, expected, error",
    [
        ('float32', 'float32', nullcontext()),
        ('f', 'float32', nullcontext()),
        ('f4', 'float32', nullcontext()),
        ('<f4', 'float32', nullcontext()),
        ('single', 'float32', nullcontext()),
        ('float64', 'float64', nullcontext()),
        ('f8', 'float64', nullcontext()),
        ('<f8', 'float64', nullcontext()),
        ('double', 'float64', nullcontext()),
        ('float', 'float64', nullcontext()),
        ('float16', '', pytest.raises(TypeError)),
        (ops.float32, 'float32', nullcontext()),
        (ops.single, 'float32', nullcontext()),
        (ops.double(10), 'float64', nullcontext()),
        (ops.float32(0), 'float32', nullcontext()),
        (float, 'float64', nullcontext()),
        (ops.dtype, '', pytest.raises(TypeError)),
        (ops.dtype(np.float32), 'float32', nullcontext()),
        (np.dtype(ops.double(10)), 'float64', nullcontext()),
        (ops.ones(10, 'float32'), 'float32', nullcontext()),
        (ops.ones(10, float).dtype, 'float64', nullcontext()),
        ('>f4', '', pytest.raises(TypeError)),
        (ops.dtype('>f8'), '', pytest.raises(TypeError)),
        (np.int32, '', pytest.raises(TypeError)),
        (ops.dtype(np.half), '', pytest.raises(TypeError)),
        ('d', 'float64', nullcontext()),
        (ops.dtype('d'), 'float64', nullcontext())
    ]
)
def test_parse_datatype(dtype, expected, error) -> None:
    with error:
        assert config.parse_datatype(dtype) == expected



"""
Defines two types of layers:
    1. Layers that are intended to be subclasses and can't be used Directly.
        1. `Layer`: It's the base class for the layers defined in this package and any user defined custom layer should
           subclass it.
        2. `SpatialLayer`: The base class of layers that deal with images like conv and pooling layers. Any custom
           layers that deal with images should subclass it because auto handles for different image formats and
           parameter checking.
        3. `Pooling2D`: Subclasses `SpatialLayer` and is the base class of `MaxPool2D` and `AvgPool2D`. It implements
           both of the operations and can be used directly by passing the `mode` argument to it. However, it's preferred
           to use `MaxPool2D` or `AvgPool2D` directly.
    2. Public layers that can be used directly.
        1. `Conv2D`.
        2. `MaxPooling`.
        3. `AvgPooling`.
        4. `BatchNormalization`.
        5. `Flatten`.
        6. `Dense`.
        7. `Dropout`.
"""
from functools import partial
from typing import Union, Callable, Optional, List, TypeVar

from xrnn import c_layers
from xrnn import config
from xrnn import layer_utils
from xrnn import ops

AnyLayer = TypeVar('AnyLayer', bound='Layer')


class Layer:

    def __init__(
            self,
            weight_l2: float = 0.,
            bias_l2: float = 0.,
            weight_initializer: str = 'auto',
            bias_initializer: str = 'zeros') -> None:
        """
        Base layer that all other layers are derived from. It has a collection of methods and interfaces that are used
        across all types of layers.

        Parameters
        ----------
        weight_l2: float
            L2 regularization value for weights.
        bias_l2: float
            L2 regularization value for biases.
        weight_initializer: str
            How to initialize the weights. Allowed methods are: 'zeros', 'ones',
            'standard_normal', 'xavier', 'he', 'auto'. *Note* when `auto` is chosen, the activation function name must
            be passed to the `build` method, if not, `xavier` initialization method is used. Available activation
            function (names) are: 'relu', 'sigmoid', 'tanh' and 'softmax'.
        bias_initializer: str
            How to initialize the biases. The Default method is an array full of zeros. Support the same arguments as
            `weight_initializer`.

        Notes
        -----
        Any custom layer should subclass Layer and must define the following:
         1. build (method): Method initializing/creating the layer weights and biases. Only needed if the layer has
            weights and/or biases.
         2. built (attribute): Set to False upon layer creation. Only needed if the layer can be built, meaning it has
            weights and/or biases, and then set to True after the layer is built by calling `build`.
         3. compute_output_shape (method): Only needed if the layer alters the shape of the input (like a dense layer).
            It should return the output shape of the layer after the inputs has passed through it.
         4. weights (attribute): Only needed if the layer has them. If the layer has weights but are called a different
            name (like BatchNorm, its weights are called gamma) and you don't want to change their name explicitly to
            weights, you can make a class property called weights that returns your weights, and a property.setattr that
            is used to modify them if necessary. Look at `~Conv2D` as an example of this.
         5. biases (attribute): Same as weights.
         6. forward (method): takes a numpy array as input, performs the calculation on it, and returns a numpy array as
            an output. *Note* that this method should implement the logic of your layer only, data type casting,
            variable batch size, and saving the input array are accounted for automatically. If you need to save other
            intermediate results that will be used during the backward pass, you can check if the layer is in training
            mode by checking for `self.training` and saving them.
         7. backward (method): takes the gradients calculated w.r.t from the layer proceeding it, and should calculate
            the gradients w.r.t its weights (save it in self.d_weights) and biases (save it in self.d_biases) if it has
            them, and w.r.t to the input of the layer and return that. Note that weight and biases l2 regularization is
            already implemented in this base layer, and all you need to do if you want to support them is to call
            `self.apply_l2_gradients()` after calculating `d_weights` and `d_biases` if your layer has those.
         8. set_params and get_params (method): Optional, only needed if your layer supports parameters other than
            weights and biases, or if you need custom logic to verify that the passed params are compatible with your
            layer.
         9. expected_dims (attribute): The expected number of dimensions that the layer expects the inputs to have. Only
            needed when the layer has a restriction over the shape of the input, otherwise don't define it.
         * For a minimal working example, see `Dense` for creating a trainable layer by subclassing `Layer`, and
           `Dropout` for a non-trainable layer example.

        If your layer deals with image inputs and changes their spatial dimensions, consider subclassing `SpatialLayer`
        instead because it handles different image formats and does parameter checking. The aforementioned points apply
        to it too.
        """

        self.weight_l2 = weight_l2  # L2 regularization lambda for weights.
        self.bias_l2 = bias_l2  # L2 regularization lambda for biases.

        self.weight_initializer = weight_initializer
        self.bias_initializer = bias_initializer

        self.training = True  # Whether the layer is used for training or not (inference). When it's not used for
        # training, or in other words, just calling the layer's forward method, there's no need to save any intermediate
        # calculations because the backward method is not going to get called, thus saving memory.

        self.inputs = None  # Saving the inputs to the layer to use them later during backprop.
        self.output = None

        self.input_shape = None
        self.expected_dims = None

        self.built = None  # Set it to None and not False because some layers don't need to be built like activation
        # layers and None is used to denote that unlike False which means the layer hasn't been built yet.

        self.d_weights = None  # Derivative w.r.t weights.
        self.d_biases = None  # Derivative w.r.t biases.

        self._dtype = config.DTYPE  # Set the default datatype.
        # Add the layer instance to the created layers list to keep track/manipulate it.
        config.CREATED_OBJECTS.append(self)
        self._name = layer_utils.make_unique_name(self)

    @property
    def name(self) -> str:  # Makes the name unchangeable.
        """
        Return the name of the layer. The name of the layer consists of two parts, the layer's type and its index.

        Examples
        --------
        >>> Conv2D(16, 3).name
        'Conv2D_0'
        >>> Conv2D(16, 3).name
        'Conv2D_1'
        """
        return self._name

    @property
    def dtype(self) -> config.DtypeHint:
        """Returns a string representing the data type of the layer. e.g. 'float32'."""
        return self._dtype

    @dtype.setter
    def dtype(self, new_dtype: config.DtypeHint):
        new_dtype = config.parse_datatype(new_dtype)
        self._dtype = new_dtype
        for attribute_name in self.__dict__:
            attribute = getattr(self, attribute_name)
            if isinstance(attribute, ops.ndarray):
                if attribute.dtype not in (new_dtype, bool):
                    setattr(self, attribute_name, attribute.astype(new_dtype, 'C'))

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        """Computes the output shape of the model based on the input shape."""
        return input_shape

    @property
    def output_shape(self) -> tuple:
        """Returns the output shape of the layer"""
        if not self.input_shape:
            raise ValueError(
                "The layer hasn't been built yet thus it has no input or output shape. Please call `build()` first.")
        return self.compute_output_shape(self.input_shape)

    @property
    def units(self) -> int:
        """A unified way to return the number of nodes for different layers. for e.g. the number of `neurons` in a
        dense layer or the number of `filters/kernels` in convolution layer."""
        raise AttributeError("This method must be overridden.")

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        """Performs a forward pass through the layer. The logic of the forward pass should be implemented in this
        method. However, to pass an input to a layer, the layer should be called with the inputs `layer(inputs)` because
        it automatically casts the inputs to the dtype of the layer. This is necessary because NumPy casts the arrays to
        the higher precision dtype, known as `upcasting`, when performing any arithmatic operations on arrays with
        different dtypes. For e.g. np.ones(10, np.float64) * np.ones(10, np.float32) will result in an array of
        dtype np.float64 which is not a desired behavior. Also, it deals with variable batch sizes and caching the
        inputs."""
        raise NotImplementedError("This method must be overridden.")

    def __call__(self, inputs: ops.ndarray) -> ops.ndarray:
        """This method should be called when passing inputs through the layer."""
        if not isinstance(inputs, ops.ndarray):
            inputs = ops.array(inputs, self.dtype, order='C')
        # Check if the input shape is compatible with the layer.
        if self.expected_dims and inputs.ndim != self.expected_dims:
            raise ValueError(
                f"{self.name} expects inputs to be {self.expected_dims} dimensional. Got input with {inputs.ndim} "
                f"dimensions.")
        curr_input_shape = inputs.shape
        if self.built is False:
            self.build(curr_input_shape)
        if not self.input_shape:  # If a custom-defined layer doesn't save the input_shape passed to its build method.
            self.input_shape = curr_input_shape
        else:
            if self.input_shape[1:] != curr_input_shape[1:]:  # Variable input shape, not supported.
                raise ValueError(
                    f"Variable input shape is not supported. The layer was originally built with shape "
                    f"(batch_size, {self.input_shape[1:]}), but inputs shape is: (batch_size, {curr_input_shape[1:]}).")
            if self.input_shape[0] != curr_input_shape[0]:  # Variable batch size.
                self.input_shape = (curr_input_shape[0],) + self.input_shape[1:]
        if inputs.dtype != self.dtype:
            inputs = inputs.astype(self.dtype)
        if getattr(self, 'padding', None) == 'same':
            window_size = getattr(self, 'window_size')
            strides = getattr(self, 'strides')
            inputs = layer_utils.pad_batch(
                inputs, layer_utils.calculate_padding_on_sides(self.input_shape, window_size, strides))
        if self.training:
            self.inputs = inputs
        return self.forward(inputs)

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        """
        Performs a backward pass through the layer calculating the gradients.

        Parameters
        ----------
        d_values: numpy array
            gradients from the next layer w.r.t to its inputs.

        Returns
        -------
        d_inputs: numpy array
            The derivative w.r.t the layer's inputs.
        """
        raise NotImplementedError("This method must be overridden.")

    @staticmethod
    def get_initialization_function(method: str, activation: Optional[str] = None) -> Callable[[tuple], ops.ndarray]:
        """
        Takes the method name as a string and returns a callable that takes `input_shape` and fills it with values
        according to `method`. "xavier" and "he" are implemented following [1]_ and [2]_ respectively.

        Parameters
        ----------
        method: {'zeros', 'ones', 'standard_normal', 'xavier', 'he', 'auto'}
            A sting indicating the initialization method name. Allowed methods are: 'zeros', 'ones',
            'standard_normal', 'xavier' `(aka Glorot Uniform)`, 'he' `(aka Kaiming)`, 'auto'. If the method is set
            to 'auto', then the initialization function is determined by the activation function. If the activation is
            'relu', `he` initialization method is used, otherwise `xaviar` method is used.
        activation: {'relu', 'tanh', 'sigmoid', 'softmax'}, optional
            A string indicating the layer activation function. Only needed if `method` is set to `auto`.

        Returns
        -------
        init_func: function
            Initialization function the takes `input_shape` and returns an array of weights with the same shape.

        Raises
        ------
        ValueError
            If `method` is not one of the supported methods, if `activation` is not one of the supported activations and
            `method` is set to 'auto'.

        References
        ----------
        .. [1] https://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf
        .. [2] https://arxiv.org/pdf/1502.01852.pdf
        """

        def uniform_init(input_shape: tuple, mode: str = 'fan_avg') -> ops.ndarray:
            """Returns a numpy function to initialize the weights using a method based on `mode`.
            If `mode` is set to 'fan_in' it uses 'He` method, if it's set to 'fan_avg' it uses `Xavier` method."""
            shape = input_shape if len(input_shape) == 2 else input_shape[1:3]  # conv kernel.
            if mode == 'fan_in':
                scale = shape[0]
            elif mode == 'fan_avg':
                scale = sum(shape)
            else:  # fan_out
                scale = input_shape[1]
            limit = ops.sqrt(6. / scale)
            return ops.random.uniform(-limit, limit, input_shape)

        method = method.lower()
        activation = activation.lower() if activation else activation
        allowed_methods = ('zeros', 'ones', 'standard_normal', 'xavier', 'he', 'auto')
        if method not in allowed_methods:
            raise ValueError(f"initialization method must be one of {allowed_methods}. Got {method}.")
        if method in ('zeros', 'ones', 'standard_normal'):
            package = ops.random if method == 'standard_normal' else ops
            return getattr(package, method)
        if method == 'auto':
            if not activation or activation in ('sigmoid', 'tanh', 'softmax'):
                return partial(uniform_init, mode='fan_avg')
            if activation == 'relu':
                return partial(uniform_init, mode='fan_in')
            raise ValueError(
                f"When using `auto` initialization method, `activation` must be one of: `relu`, `tanh`, `sigmoid` or "
                f"`softmax`. Got {activation} instead.")
        if method == 'xavier':
            return partial(uniform_init, mode='fan_avg')
        if method == 'he':
            return partial(uniform_init, mode='fan_in')

    def build(self, input_shape: Union[int, tuple], activation: str = None) -> None:
        """Builds/initialises the layer weights and biases based on the specified input shape."""
        raise NotImplementedError("This method must be overriden.")

    def apply_l2_gradients(self) -> None:
        """Calculates the partial derivative (gradients) of l2 regularization w.r.t weights and biases and applies them
        to the derived weights and biases."""
        if self.weight_l2:  # Use if statement just to squeeze a bit more performance, so we don't calculate when the
            # regularization term is equal to zero.
            self.d_weights += self.weight_l2 * 2 * getattr(self, 'weights')
        if self.bias_l2:
            self.d_biases += self.bias_l2 * 2 * getattr(self, 'biases')

    def initialize_biases(self, activation: str = None) -> Callable[[tuple], ops.ndarray]:
        """A method to initialize the layer's biases. Every layer can call this method since the way biases are
        initialized is the same across all layers."""
        return self.get_initialization_function(self.bias_initializer, activation)((self.units,))

    def get_params(self, copy: Optional[bool] = True) -> List[ops.ndarray]:
        """
        Returns a list containing the layer weights and biases.

        Parameters
        ----------
        copy: bool, optional
            Whether to copy the returned parameters. If False, any user changes to the weights and biases will
            affect the layer's weights and biases, so it's preferred to keep `copy` set to True to avoid any unintended
            behaviour caused by user code. Defaults to True

        Returns
        -------
        params: list
            A tuple containing (weights, biases, moving_mean, moving_var).

        Raises
        ------
        ValueError
            If the method was called on a layer that doesn't have weights or biases like a `Flatten` layer, if
            `get_params()` is called before the layer was built.
        """
        if self.built is None:
            raise TypeError(f"Layer of type {type(self)} doesn't have weights nor biases.")
        if self.built is False:
            raise ValueError("The layer must be built before calling `get_params()`.")
        params = [getattr(self, 'weights'), getattr(self, 'biases')]
        if copy:
            params = list(map(ops.copy, params))
        return params

    def set_params(self, params: Union[tuple, list]) -> None:
        """
        Sets the layer's parameters (weights and biases) to the new params.

        Parameters
        ----------
        params: list, tuple
            List of parameters to set, can have one element to set only the weights, or two to set weights and biases.

        Raises
        ------
        ValueError
            If one or more of the arrays in `params` shape doesn't match its respective layer parameter array shape, if
            the function was called before the layer was built, if `params` is empty.
        TypeError
            If `params` isn't a tuple or list, if the layer doesn't have weights or biases like MaxPooling2D, if the
            elements of `params` aren't numpy arrays.
        """
        if self.built is None:
            raise TypeError(f"Layer of type {type(self)} doesn't have weights nor biases.")
        if self.built is False:
            raise ValueError('The layer must be built before calling `set_params()`.')
        if not isinstance(params, (list, tuple)):
            raise TypeError(f"`params` must be a list or tuple. Got {type(params)}")
        if not len(params):
            raise ValueError('`params` can not be an empty list, it must contain at least one element.')
        for layer_param, new_param in zip(('weights', 'biases'), params):
            if not isinstance(new_param, ops.ndarray):
                raise TypeError(f"`params` contents must be numpy arrays. Got {type(new_param)}.")
            if getattr(self, layer_param).shape != new_param.shape:
                raise ValueError(
                    f"Shape mismatch for '{layer_param}', current shape is {getattr(self, layer_param).shape}, got "
                    f"{new_param.shape} new shape.")
            setattr(self, layer_param, new_param)

    def __repr__(self) -> str:
        return self.name

    def get_config(self) -> dict:
        """Return layer configuration. A layer's configuration contains the dtype of the layer, its training mode,
        dtype, input shape if defined, and the arguments used to create it."""
        layer_config = {'type': type(self).__name__, 'training': self.training, 'dtype': self.dtype}
        if self.input_shape:
            layer_config.update({'input_shape': self.input_shape})
        if self.built is not None:
            layer_config.update({
                'weight_l2': self.weight_l2,
                'bias_l2': self.bias_l2,
                'weight_initializer': self.weight_initializer,
                'bias_initializer': self.bias_initializer})
        return layer_config

    @classmethod
    def from_config(cls, layer_config: dict, build_if_can: Optional[bool] = False) -> AnyLayer:
        """
        Creates an instance of the given layer from the given config.

        Parameters
        ----------
        layer_config: dict
            A dictionary containing the layer's configuration, usually obtained from `layer.get_config()`.
        build_if_can: bool, optional
            Whether to build the created layer if input_shape was defined in the config. Defaults to False.

        Returns
        -------
        layer: Layer
            The created layer instance.

        Raises
        ------
        TypeError
            If the layer's type in `layer_config` doesn't match the layer class `from_config()` was called on, for e.g.,
            layer_config = {'type': 'Dense', ...}; layers.Conv2D.from_config(layer_config).
        """
        layer_config = layer_config.copy()
        layer_type = layer_config.pop('type')
        if layer_type != cls.__name__:
            raise TypeError(
                f"Layer type in config is {layer_type}, but `from_config()` was called on a layer of type"
                f"{cls.__name__}.")
        dtype = layer_config.pop('dtype')
        training = layer_config.pop('training')
        input_shape = layer_config.pop('input_shape', None)
        input_shape = tuple(input_shape) if input_shape else None # noqa: No PyCharm it's not None, we checked for that.
        layer = cls(**layer_config)
        if build_if_can and input_shape:
            if layer.built is not None:
                layer.build(input_shape)
            else:
                layer.input_shape = input_shape
        layer.training = training
        layer.dtype = dtype
        return layer


class Dense(Layer):

    def __init__(
            self,
            neurons: int,
            input_dim: int = None,
            weight_l2: float = 0.,
            bias_l2: float = 0.,
            weight_initializer: str = 'auto',
            bias_initializer: str = 'zeros') -> None:
        """
        A fully connected layer.

        Parameters
        ----------
        neurons: int
            Number of neurons in this layer.
        input_dim: int, optional
            Number of input features or number of the previous layer neurons.
        weight_l2: float, optional
            L2 regularization value for weights.
        bias_l2: float, optional
            L2 regularization value biases.
        weight_initializer: str, optional
            How to initialize the weights. See Layer documentation for more information.
        bias_initializer: str, optional
            How to initialize the biases. Default method is an array full of zeros.
        """
        super().__init__(weight_l2, bias_l2, weight_initializer, bias_initializer)
        self.neurons = neurons

        self.expected_dims = 2
        self.weights = None
        self.biases = None
        self.built = False
        self.input_dim = input_dim
        if input_dim:
            if self.weight_initializer != 'auto':
                self.build(input_dim)

    def build(self, input_shape: Optional[Union[int, tuple]] = None, activation: Optional[str] = None) -> None:
        if self.built:
            raise ValueError('A layer can only be built once.')
        if input_shape:  # Prioritize input_shape passed to build.
            input_d = input_shape[-1] if isinstance(input_shape, tuple) else input_shape
        else:
            if not self.input_dim:
                raise ValueError(
                    'Either `input_dim` passed to the constructor or `input_shape` passed to `build` must be provided.')
            input_d = self.input_dim
        self.weights = self.get_initialization_function(self.weight_initializer, activation)((input_d, self.units))
        self.biases = self.initialize_biases(activation)
        self.built = True
        self.input_shape = input_shape

    @property
    def units(self) -> int:
        return self.neurons

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        if not isinstance(input_shape, tuple):
            raise ValueError("`input_shape` must be a tuple of shape (batch_size, input_features).")
        return input_shape[0], self.units

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        output = ops.dot(inputs, self.weights) + self.biases[ops.newaxis,]
        return output

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        self.d_weights = ops.dot(self.inputs.T, d_values)
        self.d_biases = ops.sum(d_values, axis=0)
        self.apply_l2_gradients()
        return ops.dot(d_values, self.weights.T)

    def get_config(self) -> dict:
        layer_config = super().get_config()
        layer_config.update({'neurons': self.units})
        return layer_config


class Dropout(Layer):

    def __init__(self, rate: float) -> None:
        """
        A layer that disables up to `rate` neurons randomly. The disabled neurons are not the same each run,
        thus `randomly` disabling neurons.

        .. note:: The neurons that get "turned off" are random for each batch.

        Parameters
        ----------
        rate: float
            Percentage of neurons to disable. This value should be between zero and 1.
        """
        super().__init__()
        if not 0 <= rate <= 1:
            raise ValueError('Dropout rate value must be between zero and one.')
        self.rate = rate
        self.binary_mask = None

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        if self.training and self.rate != 0:
            # numpy.binomial takes the success rate (result equals 1), that's why we invert the rate.
            self.binary_mask = ops.random.binomial(1, 1 - self.rate, self.input_shape) / (1 - self.rate)
            return inputs * self.binary_mask
        return inputs

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        return d_values * self.binary_mask

    def get_config(self) -> dict:
        layer_config = super().get_config()
        layer_config.update({'rate': self.rate})
        return layer_config


class SpatialLayer(Layer):

    def __init__(
            self,
            window_size: Union[int, tuple],
            strides: Union[int, tuple] = 1,
            padding: config.Literal['same', 'valid'] = 'valid',
            weight_l2: float = 0.,
            bias_l2: float = 0.,
            kernel_initializer: str = 'auto',
            bias_initializer: str = 'zeros') -> None:
        """Base class for layers that alter the spatial dimensions of the input like pooling and convolution layers."""
        super().__init__(weight_l2, bias_l2, kernel_initializer, bias_initializer)
        self.window_size = layer_utils.to_tuple(window_size)
        self.strides = layer_utils.to_tuple(strides)
        self.padding = layer_utils.validate_padding(padding)

        self.expected_dims = 4
        self._padding_dims = None

    def calculate_padding_amount(self, input_shape: tuple) -> tuple:
        """Returns a tuple containing how many pixels to add on each size (pad_top, pad_bot, pad_left, pad_right)."""
        return (0, 0, 0, 0) if self.padding == 'valid' else layer_utils.calculate_padding_on_sides(
            input_shape, self.window_size, self.strides)

    @property
    def padding_amount(self) -> tuple:
        """Returns a tuple containing how many pixels to add on each size (pad_top, pad_bot, pad_left, pad_right)."""
        if not self._padding_dims:
            self._padding_dims = self.calculate_padding_amount(self.input_shape)
        return self._padding_dims

    @property
    def padded_input_shape(self) -> tuple:
        """Returns the shape of the padded inputs."""
        return layer_utils.padded_input_shape(self.input_shape, self.padding_amount)

    @property
    def nhwc(self) -> bool:
        """Returns True if the current image data format is NHWC (channels-last) and False if NCHW (channels-first)"""
        return config.IMAGE_DATA_FORMAT == 'channels-last'

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        padding_dims = self.calculate_padding_amount(input_shape)
        height, width = layer_utils.compute_spatial_output_shape(
            input_shape, self.window_size, self.strides, padding_dims)
        out_channels = getattr(self, 'units', None)
        if self.nhwc:
            output_shape = (input_shape[0], height, width, out_channels if out_channels else input_shape[-1])
        else:
            output_shape = (input_shape[0], out_channels if out_channels else input_shape[1], height, width)
        return output_shape

    def to_nhwc_format(self, shape: tuple) -> tuple:
        """Returns the shape in NHWC format if it's in NCHW format only, otherwise return it unchanged."""
        if not self.nhwc:
            return shape[0], shape[2], shape[3], shape[1]
        return shape

    def make_arguments_list(self, *args: ops.ndarray) -> tuple:
        """
        Adds specific arguments after the provided *args that are passed universally to all convolution and pooling
        operations (forward and backward).

        Parameters
        ----------
        args: tuple or list of numpy arrays
            A container of numpy arrays, the provided args are going to be placed at the start of the returned complete
            argument list.

        Returns
        -------
        complete_args_list: tuple
            The complete argument list containing *args at the start and the rest is the universal input parameters of
            pooling and convolution c functions.
        """
        end_idx = 3 if issubclass(self.__class__, Pooling2D) else 4
        return (*args,
                *self.window_size, *self.strides,
                *self.to_nhwc_format(self.output_shape),
                *self.to_nhwc_format(self.padded_input_shape)[1:end_idx],
                self.nhwc)

    def get_config(self) -> dict:
        layer_config = super().get_config()
        window_key = 'kernel_size' if isinstance(self, Conv2D) else 'pool_size'
        layer_config.update({window_key: self.window_size, 'strides': self.strides, 'padding': self.padding})
        kernel_initializer = layer_config.pop('weight_initializer', None)
        if kernel_initializer:
            layer_config.update({'kernel_initializer': kernel_initializer})
        return layer_config


class Conv2D(SpatialLayer):

    def __init__(
            self,
            kernels: int,
            kernel_size: Union[int, tuple],
            strides: Union[int, tuple] = 1,
            padding: config.Literal['same', 'valid'] = 'valid',
            weight_l2: float = 0.,
            bias_l2: float = 0.,
            kernel_initializer: str = 'auto',
            bias_initializer: str = 'zeros') -> None:
        """
        A convolution 2D Layer.

        Parameters
        ----------
        kernels: int
            The number of kernels/filters to use.
        kernel_size : int or tuple
            An int or tuple specifying the kernel sizes. In the case of an int, the same dimension is used for both the
            height and width.
        strides: int or tuple, optional
            An int or tuple specifying the strides along the width and height of the kernels. Default is 1.
        padding: str, optional
            'same' or 'valid'. 'same' zero pads the input evenly. 'valid' means no padding. Default is 'valid'.
        weight_l2: float, optional
            L2 regularization value for weights. Default is 0.
        bias_l2: float, optional
            L2 regularization value for biases. Default is 0.
        kernel_initializer: str, optional
            How to initialize the kernels. See Layer documentation for more information. Default is 'auto'.
        bias_initializer: str, optional
            How to initialize the biases. Default method is an array full of zeros.
        """
        super().__init__(kernel_size, strides, padding, weight_l2, bias_l2, kernel_initializer, bias_initializer)
        self.n_kernels = kernels

        self.kernels = None  # Kernels are of shape (n_filters, kernel_height, kernel_width, n_channels).
        self.biases = None
        self.built = False

    @property
    def weights(self) -> ops.ndarray:
        """A property that returns the kernels of the layer which is the same as weights but is named kernels because
        that's more convenient. This property is needed because different parts of the package use and manipulate the
        weights attribute of the layer and this layer doesn't explicitly have it, so this method makes all the calls
        consistent across layers."""
        return self.kernels

    @weights.setter
    def weights(self, new_weights) -> None:
        self.kernels = new_weights

    @property
    def units(self) -> int:
        return self.n_kernels

    def build(self, input_shape: tuple, activation: str = None) -> None:
        if self.built:
            raise ValueError('A layer can only be built once.')
        if len(input_shape) not in (3, 4):
            raise ValueError(
                f"`input_shape` must have a length of 3 (not including batch dimension) or 4 "
                f"(including batch dimension). Got {input_shape}")
        if len(input_shape) == 3:
            input_shape = (1, *input_shape)
        channels = input_shape[-1] if self.nhwc else input_shape[1]
        self.kernels = self.get_initialization_function(self.weight_initializer, activation)(
            (self.n_kernels, *self.window_size, channels))
        self.biases = self.initialize_biases(activation)
        self.built = True
        self.input_shape = input_shape

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        feature_maps = ops.zeros(self.output_shape)
        args = self.make_arguments_list(inputs, self.kernels, self.biases, feature_maps)
        c_layers.convForwardF(*args) if self.dtype == 'float32' else c_layers.convForwardD(*args)
        return feature_maps

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        d_inputs = ops.zeros(self.padded_input_shape)
        self.d_weights = ops.zeros(self.kernels.shape)
        self.d_biases = d_values.sum((0, 1, 2) if self.nhwc else (0, 2, 3))
        args = self.make_arguments_list(self.inputs, self.kernels, d_values, self.d_weights, d_inputs)
        c_layers.convBackwardF(*args) if self.dtype == 'float32' else c_layers.convBackwardD(*args)
        d_inputs = layer_utils.extract_from_padded(d_inputs, self.padding_amount)
        self.apply_l2_gradients()
        return d_inputs

    def set_params(self, params: Union[list, tuple], keras_weights: Optional[bool] = False) -> None:
        """
        Sets the layer's parameters (weights and biases) to the new params.

        Parameters
        ----------
        params: list, tuple
            List of parameters to set, can have one element to set only the weights, or two to set weights and biases.
        keras_weights: bool, optional
            Whether the weights came from a keras layer. This needs to be explicitly satated because keras Conv2D layer
            weights' (kernel) has a different shape from this layer weights' (kernel). Defaults to False

        Raises
        ------
        ValueError
            If one or more of the arrays in `params` shape doesn't match its respective layer parameter array shape, if
            the function was called before the layer was built.
        TypeError
            If `params` isn't a tuple or list.
        """
        if keras_weights:
            params = [params[0].transpose(3, 0, 1, 2)]
            if len(params) == 2:
                params.append(params[1])
        super().set_params(params)

    def get_config(self) -> dict:
        layer_config = super().get_config()
        layer_config.update({'kernels': self.n_kernels})
        return layer_config


class Pooling2D(SpatialLayer):

    def __init__(
            self,
            pool_size: Union[tuple, int],
            strides: Optional[Union[tuple, int]] = 1,
            padding: config.Literal['same', 'valid'] = 'valid',
            mode: config.Literal['max', 'avg'] = None) -> None:
        """Base class for MaxPooling2D and AvgPooling2D. The implementation of both is almost identical, the only
        Difference is how the results are calculated, which can be dictated by `mode` argument, 'max' or 'avg'."""
        super().__init__(pool_size, strides, padding)
        if mode not in ('max', 'avg'):
            raise ValueError(f"Mode must either be 'max' or 'avg'. Got '{mode}'")
        self.masks = None
        self.use_max = mode == 'max'
        self.expected_dims = 4

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        feature_maps = ops.zeros(self.output_shape)
        self.masks = ops.zeros(self.padded_input_shape)
        args = (inputs, self.masks, feature_maps) if self.use_max else (inputs, feature_maps)
        args = self.make_arguments_list(*args)
        if self.dtype == 'float32':
            c_layers.maxPoolForwardF(*args) if self.use_max else c_layers.avgPoolForwardF(*args)
        else:
            c_layers.maxPoolForwardD(*args) if self.use_max else c_layers.avgPoolForwardD(*args)
        return feature_maps

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        d_inputs = ops.zeros(self.padded_input_shape)
        args = (d_values, self.masks, d_inputs) if self.use_max else (d_values, d_inputs)
        args = self.make_arguments_list(*args)
        if self.dtype == 'float32':
            c_layers.maxPoolBackwardF(*args) if self.use_max else c_layers.avgPoolBackwardF(*args)
        else:
            c_layers.maxPoolBackwardD(*args) if self.use_max else c_layers.avgPoolBackwardD(*args)
        d_inputs = layer_utils.extract_from_padded(d_inputs, self.padding_amount)
        return d_inputs


class MaxPooling2D(Pooling2D):
    def __init__(
            self,
            pool_size: Union[tuple, int],
            strides: Optional[Union[tuple, int]] = 1,
            padding: config.Literal['same', 'valid'] = 'valid') -> None:
        """
        Max pooling operation for 2D spatial data like images. It performs a max operation over a window for each
        channel of the inputs, down sampling the spatial dimensions in the process.

        Parameters
        ----------
        pool_size: int, tuple
            The window height and width to perform the max operation on. If integer, it's used for both window height
            and width.
        strides: int, tuple, optional
            How far the pooling window moves each step. If integer, the window will slide by this factor for height and
            width. Default is 1.
        padding: str, optional
            Either 'same' or 'valid'. 'valid' applies no padding to the input. 'same' evenly pads the spatial dimensions
            of the input such that the output has the same height and width when `strides=1`, otherwise spatial
            dimensions // strides. Default is valid
        """
        super().__init__(pool_size, strides, padding, 'max')


class AvgPooling2D(Pooling2D):
    def __init__(
            self,
            pool_size: Union[tuple, int],
            strides: Optional[Union[tuple, int]] = 1,
            padding: config.Literal['same', 'valid'] = 'valid') -> None:
        """
        Average pooling operation for 2D spatial data like images. It downsamples the input by taking the average over a
        window (dictated by `pool_size`) for each input channel.

        Parameters
        ----------
        pool_size: int, tuple
            The window height and width to take the average over. If integer, it's used for both window height
            and width.
        strides: int, tuple, optional
            How far the pooling window moves each step. If integer, the window will slide by this factor for height and
            width. Default is 1.
        padding: str, optional
            Either 'same' or 'valid'. 'valid' applies no padding to the input. 'same' evenly pads the spatial dimensions
            of the input such that the output has the same height and width when `strides=1`, otherwise spatial
            dimensions // strides. Default is valid
        """
        super().__init__(pool_size, strides, padding, 'avg')


class BatchNormalization(Layer):

    def __init__(
            self,
            axis: Union[int, tuple, list] = None,
            momentum: float = 0.9,
            gamma_initializer: str = 'ones',
            beta_initializer: str = 'zeros',
            gamma_l2: float = 0.,
            beta_l2: float = 0.) -> None:
        """
        Normalizes the layer inputs by keeping the mean close to zero the standard deviation close to one.

        Parameters
        ----------
        axis : int, tuple, list, optional
            The axis to be normalized, typically the channels' axis. For instance, when the input data format is
            'channels-last', the axis can be 3 or -1. The default value `None` means to infer the axis automatically
            based on the input images data format. Can be a list/tuple of ints to specify multiple axes.
        momentum : float, optional
            Momentum for the moving average. Think of it as the weight of the previous observations. Default is 0.9.
        gamma_initializer : str, optional
            Initializer for gamma weights. Default is 'ones'.
        beta_initializer : str, optional
            Initializer for beta weights. Default is 'zeros'.
        gamma_l2 : float, optional
            L2 regularization value for gamma weights. Default is 0.
        beta_l2 : float, optional
            L2 regularization value for beta weights. Default is 0.
        """
        super().__init__(gamma_l2, beta_l2, gamma_initializer, beta_initializer)
        self.axis = axis
        self.momentum = momentum

        self.reduction_axis = []
        # Save these variables when the layer is in training mode because they are first calculated in the forward pass
        # and then used in the backward pass, so we save them to avoid recalculating them during the backward pass.
        self.stddev = None
        self.moving_mean = None
        self.moving_var = None
        self.normalized_x = None
        self.weights = None  # Gamma.
        self.biases = None  # Beta.
        self.built = False

    def get_reduction_axis(self, n_dims: int) -> tuple:
        """Returns a tuple of the axis(es) to perform the mean and variance calculation on. For instance, the given axis
        is -1, and data format is `channels-last` this gives (0, 1, 2) reduction axis.

        Parameters
        ----------
        n_dims: int
            How many dimensions that the input is expected to have, for e.g., it should be 4 for image inputs.
        """
        if n_dims not in (2, 4):
            raise ValueError(f"`n_dims` must be either 2 or 4. Got {n_dims}")
        if isinstance(self.axis, (list, tuple)) and len(self.axis) == n_dims:
            raise ValueError(
                "`axis` can't have the same number of axes as `n_dims`, this will result in reducing the output to a "
                "scaler value, which isn't supported and probably not what you intended.")
        if not self.axis:
            self.axis = -1 if config.IMAGE_DATA_FORMAT == 'channels-last' or n_dims == 2 else 1
        if not isinstance(self.axis, list):
            # Convert it to a list to be able to assign values to it (tuples are mutable, and ints are not iterable).
            if isinstance(self.axis, int):
                self.axis = [self.axis]
            else:
                self.axis = list(self.axis)
        for i in range(len(self.axis)):
            if self.axis[i] < 0:  # Handle end-relative (negative) axis.
                self.axis[i] = n_dims + self.axis[i]  # Get absolute axis.
        return tuple([i for i in range(n_dims) if i not in self.axis])

    def build(self, input_shape: Union[int, tuple], activation: str = None) -> None:
        if self.built:
            raise ValueError('A layer can only be built once.')
        if isinstance(input_shape, int):
            input_shape = (1, input_shape)
        elif len(input_shape) == 3:
            input_shape = (1, *input_shape)
        if len(input_shape) not in (2, 4):
            raise ValueError(f'`input_shape` must have a length of 2 or 4. Got {input_shape}')
        self.reduction_axis = self.get_reduction_axis(len(input_shape))
        arrays_shape = ops.array(input_shape)
        arrays_shape[list(self.reduction_axis)] = 1
        self.weights = self.get_initialization_function(self.weight_initializer, activation)(tuple(arrays_shape))
        self.biases = self.get_initialization_function(self.bias_initializer, activation)(tuple(arrays_shape))
        self.moving_mean = ops.zeros(arrays_shape)
        self.moving_var = ops.ones(arrays_shape)
        self.built = True
        self.input_shape = input_shape

    def calculate_moving_average(self, moving: ops.ndarray, new_sample: ops.ndarray) -> ops.ndarray:
        """Calculates the moving average and updates it."""
        return self.momentum * moving + (1 - self.momentum) * new_sample

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        if self.training:
            mean = inputs.mean(self.reduction_axis, keepdims=True)
            variance = inputs.var(self.reduction_axis, keepdims=True)
            self.moving_mean = self.calculate_moving_average(self.moving_mean, mean)
            self.moving_var = self.calculate_moving_average(self.moving_var, variance)
        else:
            mean = self.moving_mean
            variance = self.moving_var
        stddev = ops.sqrt(variance + config.EPSILON)
        normalized = (inputs - mean) / stddev
        if self.training:
            # Cache these calculation because they are used in the backward pass to save computational resources.
            self.stddev = stddev
            self.normalized_x = normalized
        return self.weights * normalized + self.biases

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        n = d_values.shape[0]  # Number of samples in each batch.
        d_norm = d_values * self.weights
        self.d_weights = ops.sum(d_values * self.normalized_x, self.reduction_axis, keepdims=True)
        self.d_biases = ops.sum(d_values, self.reduction_axis, keepdims=True)
        self.apply_l2_gradients()
        return (1 / n) * (1 / self.stddev) * (
                n * d_norm - ops.sum(d_norm, self.reduction_axis, keepdims=True) -
                self.normalized_x * ops.sum(d_norm * self.normalized_x, self.reduction_axis, keepdims=True))

    def get_params(self, copy: Optional[bool] = True, squeeze: Optional[bool] = False) -> List[ops.ndarray]:
        """
        Returns a list containing the weights (gamma), biases (beta), moving mean and moving variance arrays
        respectively.

        Parameters
        ----------
        copy: bool, optional
            Whether to copy the returned parameters. If False, any user changes to the weights and biases will
            affect the layer's weights and biases, so it's preferred to keep `copy` set to True to avoid any unintended
            behaviour caused by user code. Defaults to True
        squeeze: bool, optional
            Whether to remove excess axes of length 1. This might be useful if you want to pass the parameters to a
            keras layer since keras BatchNorm parameters don't have extra empty dimensions. Defaults to False

        Returns
        -------
        params: list
            A tuple containing (weights, biases, moving_mean, moving_var).

        Raises
        ------
        ValueError
            If `get_params()` is called before the layer is built.
        """
        if not self.built:
            raise ValueError("The layer must be built before calling `get_params()`.")
        params = [self.weights, self.biases, self.moving_mean, self.moving_var]
        if copy:
            params = list(map(ops.copy, params))
        if squeeze and self.weights is not None and self.weights.shape.count(1) == 3:  # Only remove (1) dims
            params = list(map(ops.squeeze, params))
        return params

    def set_params(self, params: Union[tuple, list]) -> None:
        """
        Sets the layer's parameters using `params`. The layer must be built before calling this method, either by
        calling the layer `layer(inputs)` or by calling `layer.build(input_shape)`

        Parameters
        ----------
        params: tuple, list
            A tuple containing (weights, biases, moving_mean, moving_var). Alternatively, you can pass a subset of
            the parameters, for e.g., (weights, biases), will only set the layer weights' and biases'.

        Raises
        ------
        ValueError
            If one or more of the arrays in `params` shape doesn't match its respective layer parameter array shape, if
            the function was called before the layer was built, if `params` is empty.
        TypeError
            If `params` isn't a tuple or list, if the elements of `params` aren't numpy arrays.
        """
        if not self.built:
            raise ValueError('The layer must be built before calling `set_params()`.')
        if not isinstance(params, (list, tuple)):
            raise TypeError(f"`params` must be a list or tuple. Got {type(params)}")
        if not len(params):
            raise ValueError('`params` can not be an empty list, it must contain at least one element.')
        for layer_param, new_param in zip(('weights', 'biases', 'moving_mean', 'moving_var'), params):
            if not isinstance(new_param, ops.ndarray):
                raise TypeError(f"`params` contents must be numpy arrays. Got {type(new_param)}.")
            if new_param.ndim == 1:
                new_param = ops.expand_dims(new_param, list(ops.where(ops.array(self.weights.shape) == 1)[0]))
            if getattr(self, layer_param).shape != new_param.shape:
                raise ValueError(
                    f"Shape mismatch for '{layer_param}', current shape is {getattr(self, layer_param).shape}, got "
                    f"{new_param.shape} new shape")
            setattr(self, layer_param, new_param)

    def get_config(self) -> dict:
        layer_config = super().get_config()
        layer_config.update({'axis': self.axis, 'momentum': self.momentum})
        for key in layer_config.copy():
            val = layer_config.pop(key)
            key = key.replace('weight', 'gamma')
            key = key.replace('bias', 'beta')
            layer_config[key] = val
        return layer_config


class Flatten(Layer):
    """
    Flattens the inputs across feature dimensions and keeps the batch dimension.

    Examples
    --------
    >>> flatten = Flatten()
    >>> batch = ops.ones((256, 28, 28, 3))
    >>> output = flatten.forward(batch)
    >>> output.shape
    (256, 2352)
    """

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        return input_shape[0], ops.prod(input_shape[1:]).item()

    def forward(self, inputs: ops.ndarray) -> ops.ndarray:
        return inputs.reshape((inputs.shape[0], -1))

    def backward(self, d_values: ops.ndarray) -> ops.ndarray:
        return d_values.reshape(self.input_shape)



"""Implements Adam, SGD, RMSprop, Adagrad optimization algorithms. Also implements the base Optimizer class the custom
optimizers should subclass, see `~Optimizer` docstrings for more info on subclassing behaviour."""
from typing import Callable, TypeVar

from xrnn import config
from xrnn import ops

DecayFunSig = Callable[[float, float, int, int], float]
AnyOptimizer = TypeVar('AnyOptimizer', bound='Optimizer')


class Optimizer:

    def __init__(
            self,
            learning_rate: float = 0.001,
            decay: float = 0.,
            decay_func: DecayFunSig = None) -> None:
        """
        Base optimizer class. Implements universal methods used by all other optimizers like handling learning rate
        decay and keeping track of some variables.

        Parameters
        ----------
        learning_rate: float, optional
            The learning rate value. Each optimizer has a different default learning rate value.
        decay: float, optional
            Set this value to a number greater than zero if you want to exponentially decay the learning rate per
            iteration (batch) according to the following: `initial_lr * (1 / (1 + decay_value * current_iteration))`.
        decay_func: Callable, optional
            A custom decaying function. This function must accept four arguments in this order:
            `initial_lr`, `current_lr`, `iteration`, `epoch`. And it should return a float/int value which is the new
            lr. This function is called every step/iteration, if your function decays based on epochs just check for the
            passed `epoch` argument.

        Raises
        ------
        ValueError
            If both `decay` and `decay_func` are both set.

        Notes
        -----
        Custom optimizers must subclass this class and implement the following.

        * `update_params (method)`: Updates the weights and biases of the optimizer based on the optimization
          algorithm. Layer's weights and biases are attributes of each layer (layer.weights, layer.biases) and weights
          and biases gradients are also layer attributes (layer.d_weights, layer.d_biases). `pre_update_params` method
          must be called at the top of the method (`update_params`) definition because it performs any necessary
          initialization and updates the learning rate if necessary. if your optimization algorithm performs an
          intermediate calculation on the weights or biases and needs to save it, pass `cache=True` to
          `pre_update_params` and save the intermediates to `layer.weights_cache` and `layer.biases_cache`. If your
          optimization algorithm calculates weight or biases momentums like SGD with momentum, pass `momentums=True` to
          `pre_update_params` and save the results to `layer.weight_momentums` and `layer.bias_momentums`.
        """

        self.learning_rate = learning_rate
        self.decay = decay
        self.decay_func = decay_func

        self.current_lr = learning_rate
        self.iterations = 0
        self.epoch = 0

        if decay != 0. and decay_func:
            raise ValueError("You can't set both `decay` and `decay_fun`, please just set one.")
        if self.decay_func:
            self.validate_decay_function(decay_func)

    @staticmethod
    def initialize_layer_state(layer, cache: bool = True, momentums: bool = True) -> None:
        """
        Some optimizers keep a cache of previously calculated gradients. This method initialises placeholders that
        are going to be used by the optimizer when updating the layer parameters'.

        .. note::
           Weight/bias cache and weight/bias momentums cache are bound to the layer not the optimizer, so if a new
           optimizer is chosen for the same model, it will use the previously declared and used caches.

        Parameters
        ----------
        layer: Layer
            An instance of `xrnn.layers.Layer`.
        cache: bool, optional
            Whether to create cache arrays for the weights and biases.
        momentums: bool, optional
            Whether to create momentum arrays for the weights and biases.
        """
        if not hasattr(layer, 'weights_cache') and cache:
            layer.weights_cache = ops.zeros(layer.weights.shape, layer.weights.dtype)
            layer.biases_cache = ops.zeros(layer.biases.shape, layer.biases.dtype)
        if not hasattr(layer, 'weight_momentums') and momentums:
            layer.weight_momentums = ops.zeros(layer.weights.shape, layer.weights.dtype)
            layer.bias_momentums = ops.zeros(layer.biases.shape, layer.biases.dtype)

    @staticmethod
    def validate_decay_function(decay_func: DecayFunSig) -> None:
        """Checks if the provided custom `decay_func` can be used."""
        try:
            new_lr = decay_func(0.1, 1e-3, 0, 0)
            if not isinstance(new_lr, (float, int)):
                raise TypeError(f"`decay_func` should return an `int` or `float` which is the new learning rate. "
                                f"Got {type(new_lr)}.")
        except Exception as e:
            if '`decay_func`' in str(e):
                raise e
            raise RuntimeError("`decay_func` didn't work and raised the previous exception") from e

    def update_learning_rate(self) -> None:
        """Updates the current learning rate if `decay` or `decay_func` is set."""
        if self.decay:
            self.current_lr = self.learning_rate * (1. / (1. + self.decay * self.iterations))
        if self.decay_func:
            self.current_lr = self.decay_func(self.learning_rate, self.current_lr, self.iterations, self.epoch)

    def pre_update_params(self, layer, cache: bool = False, momentums: bool = False) -> None:
        """
        Must be called before updating the params to set the initial state and perform any decaying method set.

        Parameters
        ----------
        layer: Layer
            An instance of `xrnn.layers.Layer`.
        cache: bool, optional
            Whether to create cache arrays for the weights and biases.
        momentums: bool, optional
            Whether to create momentum arrays for the weights and biases.
        """
        self.initialize_layer_state(layer, cache, momentums)
        self.update_learning_rate()

    def update_params(self, layer) -> None:
        """Updates a layer weights and biases based on how the optimizer calculates the parameter updates."""
        raise NotImplementedError("This method must be overridden.")

    def get_config(self) -> dict:
        """Returns the optimizer configuration. It contains the current learning rate, initial learning rate, decay
        rate, current iteration and epoch, other optimizer specific parameters like momentum for sgd for example."""
        return {
            'type': type(self).__name__,
            'learning_rate': self.learning_rate,
            'current_lr': self.current_lr,
            'decay': self.decay,
            'iterations': self.iterations,
            'epoch': self.epoch
        }

    @classmethod
    def from_config(cls, opt_config: dict) -> AnyOptimizer:
        """
        Create a new optimizer instance from a configuration dictionary.

        Raises
        ------
        TypeError
            If the optimizer's type in `opt_config` doesn't match the optimizer class `from_config()` was called on,
            for e.g, opt_config = {'type': 'Adam', ...}; optimizer.SGD.from_config(opt_config).
        """
        opt_config = opt_config.copy()
        opt_type = opt_config.pop('type')
        if opt_type != cls.__name__:
            raise TypeError(
                f"Optimizer type in config is '{opt_type}', but `from_config()` was called on optimizer of type"
                f"{cls.__name__}.")
        curr_lr = opt_config.pop('current_lr')
        iterations = opt_config.pop('iterations')
        epoch = opt_config.pop('epoch')
        opt = cls(**opt_config)
        opt.current_lr = curr_lr
        opt.iterations = iterations
        opt.epoch = epoch
        return opt


class SGD(Optimizer):

    def __init__(
            self,
            learning_rate: float = 0.01,
            momentum: float = 0.,
            decay: float = 0.,
            decay_func: DecayFunSig = None) -> None:
        """
        Stochastic Gradient Descent optimization algorithm class.

        One of the oldest optimization algorithms that was
        used to train neural networks and is still heavily in use to this day because of how effective yet simple it is,
        especially when used with momentum. (Can be achieved by setting `momentum` to a value other than zero in the
        constructor).

        Parameters
        ----------
        learning_rate: float, optional
            The learning rate value. Each optimizer has a different default learning rate value.
        momentum: float, optional
            Just like momentum in physics, this hyperparameter builds inertia in the descending direction to overcome
            local minima and oscillation of noisy gradients. Default is zero, which is vanilla gradient descent. A value
            closer to one means keep a longer history of the previous updates resulting in more momentum.
        decay: float, optional
            Set this value to a number greater than zero if you want to exponentially decay the learning rate.
        decay_func: Callable, optional
            A custom decaying function. This function must accept four arguments in this order:
            `initial_lr`, `current_lr`, `iteration`, `epoch`. And it should return a float/int value which is the new
            lr. This function is called every step/iteration, if your function decays based on epochs just check for the
            passed `epoch` argument.
        """

        super().__init__(learning_rate, decay, decay_func)
        if not 0 <= momentum <= 1:
            raise ValueError("`momentum` must be between [0, 1].")
        self.momentum = momentum

    def update_params(self, layer) -> None:
        super().pre_update_params(layer, cache=False, momentums=True)
        if self.momentum:
            weight_updates = self.momentum * layer.weight_momentums - self.current_lr * layer.d_weights
            bias_updates = self.momentum * layer.bias_momentums - self.current_lr * layer.d_biases
            layer.weight_momentums = weight_updates
            layer.bias_momentums = bias_updates
        else:
            weight_updates = -self.current_lr * layer.d_weights
            bias_updates = -self.current_lr * layer.d_biases
        layer.weights += weight_updates
        layer.biases += bias_updates

    def get_config(self) -> dict:
        opt_config = super().get_config()
        opt_config.update({'momentum': self.momentum})
        return opt_config


class Adagrad(Optimizer):
    """
    Adaptive Gradient optimization algorithm class.

    Adagrad is (in theory) well suited for dealing with sparse data.
    However, it's not used much in the real world especially for computer vision tasks.
    """

    def update_params(self, layer) -> None:
        super().pre_update_params(layer, cache=True, momentums=False)
        layer.weights_cache += ops.square(layer.d_weights)
        layer.biases_cache += ops.square(layer.d_biases)

        # Epsilon is used, so we don't encounter division by zero errors.
        layer.weights += -self.current_lr * layer.d_weights / (ops.sqrt(layer.weights_cache) + config.EPSILON)
        layer.biases += -self.current_lr * layer.d_biases / (ops.sqrt(layer.biases_cache) + config.EPSILON)


class RMSprop(Optimizer):

    def __init__(
            self,
            learning_rate: float = 1e-3,
            decay: float = 0.,
            decay_func: Callable[[float, float, int, int], float] = None,
            rho: float = 0.9) -> None:
        """
        Root Mean Squared Propagation optimization algorithm class.

        This algorithm is derived from the concepts of SGD and Adagrad optimization algorithms. It's an adaptive
        algorithm that uses an adaptive step size for each input variable using a decaying moving average of gradients.

        Parameters
        ----------
        learning_rate: float, optional
            The learning rate value. Each optimizer has a different default learning rate value.
        decay: float, optional
            Set this value to a number greater than zero if you want to exponentially decay the learning rate.
        decay_func: Callable, optional
            A custom decaying function. This function must accept four arguments in this order:
            `initial_lr`, `current_lr`, `iteration`, `epoch`. And it should return a float/int value which is the new
            lr. This function is called every step/iteration, if your function decays based on epochs just check for the
            passed `epoch` argument.
        rho: float, optional
            The gradient exponentially weighted average decay factor, in other words, how much to consider the old
            gradients. Default: 0.9
        """

        super().__init__(learning_rate, decay, decay_func)
        self.rho = rho

    def update_params(self, layer) -> None:
        super().pre_update_params(layer, cache=True, momentums=False)
        layer.weights_cache = self.rho * layer.weights_cache + (1 - self.rho) * ops.square(layer.d_weights)
        layer.biases_cache = self.rho * layer.biases_cache + (1 - self.rho) * ops.square(layer.d_biases)

        layer.weights += -self.current_lr * layer.d_weights / (ops.sqrt(layer.weights_cache) + config.EPSILON)
        layer.biases += -self.current_lr * layer.d_biases / (ops.sqrt(layer.biases_cache) + config.EPSILON)

    def get_config(self) -> dict:
        opt_config = super().get_config()
        opt_config.update({'rho': self.rho})
        return opt_config


class Adam(Optimizer):

    def __init__(
            self,
            learning_rate: float = 1e-3,
            decay: float = 0.,
            decay_func: DecayFunSig = None,
            beta_1: float = 0.9,
            beta_2: float = 0.999) -> None:
        """
        Adaptive Moment Estimation optimization algorithm class.

        Adam is an adaptive optimization algorith that is an improvement over MSGD and RMSprop algorithms and is
        designed to speed neural networks training and reach convergence quickly by taking into consideration the
        exponentially weighted average of gradients.
        Adam is widely used in real world models and research papers and often outperforms all other optimization
        algorithms in terms of convergence speed (reaches same model performance as other techniques but with less
        training time) and better model performance. Using it as the goto optimizer is common and usually a good choice.

        .. note::
           Adam is a more complex algorithm than SGD and that is reflected in its computational footprint, the
           difference is small (2-5% on a real dataset), but it's there.

        Parameters
        ----------
        learning_rate: float, optional
            The learning rate value. Each optimizer has a different default learning rate value. Default: 1e-3.
        decay: float, optional
            Set this value to a number greater than zero if you want to exponentially decay the learning rate.
        decay_func: Callable, optional
            A custom decaying function. This function must accept four arguments in this order:
            `initial_lr`, `current_lr`, `iteration`, `epoch`. And it should return a float/int value which is the new
            lr. This function is called every step/iteration, if your function decays based on epochs just check for the
            passed `epoch` argument.
        beta_1: float, optional
            The exponential decay rate for the 1st moment estimates. Default: `0.9`.
        beta_2: float, optional
            The exponential decay rate for the 2nd moment estimates. Default: `0.999`.
        """

        super().__init__(learning_rate, decay, decay_func)
        self.beta_1 = beta_1  # Same as momentum in SGD
        self.beta_2 = beta_2  # Same as rho in RMSprop

    def update_params(self, layer) -> None:
        super().pre_update_params(layer, cache=True, momentums=True)
        # Almost the same as how SGD momentum is calculated but without the learning rate. the learning rate is used
        # in the final calculation
        layer.weight_momentums = self.beta_1 * layer.weight_momentums + (1 - self.beta_1) * layer.d_weights
        layer.bias_momentums = self.beta_1 * layer.bias_momentums + (1 - self.beta_1) * layer.d_biases

        # Get corrected momentum. Weight momentum updates are multiple times bigger than their original values in the
        # initial steps of training. As training goes on, the corrected values become closer and closer to their actual
        # values because momentums / (1 - 0.9^100) = cache / 0.9999 which is almost the same as cache / 1 = cache.
        weight_momentum_corrected = layer.weight_momentums / (1 - pow(self.beta_1, self.iterations + 1))
        bias_momentum_corrected = layer.bias_momentums / (1 - pow(self.beta_1, self.iterations + 1))

        # Update cache with squared current gradients (Same as RMSprop)
        layer.weights_cache = self.beta_2 * layer.weights_cache + (1 - self.beta_2) * ops.square(layer.d_weights)
        layer.biases_cache = self.beta_2 * layer.biases_cache + (1 - self.beta_2) * ops.square(layer.d_biases)

        # Get corrected cache. Same as corrected momentum updates (But a bit slower to reach 1).
        # Big updates in the initial steps, then it slows down as training continues
        weights_cache_corrected = layer.weights_cache / (1 - pow(self.beta_2, self.iterations + 1))
        biases_cache_corrected = layer.biases_cache / (1 - pow(self.beta_2, self.iterations + 1))

        # Vanilla SGD parameter update + normalization
        # with square rooted cache (Same as Adagrad and RMSprop)
        layer.weights += -self.current_lr * weight_momentum_corrected / (
                ops.sqrt(weights_cache_corrected) + config.EPSILON)
        layer.biases += -self.current_lr * bias_momentum_corrected / (
                ops.sqrt(biases_cache_corrected) + config.EPSILON)

    def get_config(self) -> dict:
        opt_config = super().get_config()
        opt_config.update({'beta_1': self.beta_1, 'beta_2': self.beta_2})
        return opt_config



"""This module loads the shared library that contain function definitions for convolution and pooling that are written
in C and compiled into a shared library for performance reasons and makes these functions callable from Python."""
import ctypes
import os
import platform
from typing import Union, Optional, Any, Callable

from xrnn import ops


def make_argtypes_list(
        dtype: Union[type, str, ops.dtype],
        ndim: Optional[int] = None,
        shape: Optional[Union[int, tuple, list, ops.ndarray]] = None,
        flags: Optional[Union[str, tuple]] = None,
        npointers: int = 1,
        numeral_type=ctypes.c_size_t,
        n_numerals: int = 1,
        takes_bool: bool = True) -> list:
    """
    Returns a list containing all the argument types for the inputs to C function

    Parameters
    ----------
    dtype: type, str, dtype instance
        The data type of the array that the C functions take.
    ndim: int, optional
        The number of dimensions the array should be expected to have.
    shape: tuple, optional
        The shape the array is expected to have.
    flags: str, list, optional
        Any flags the array is expected to have. For example the 'C' flags checks if the array is C contagious.
    npointers: int, optional
        The number of array pointers to create with the above checks/restrictions. For example if the C function expects
        4 arrays as inputs, set this to 4.
    numeral_type: int
        The data type that the C function expects the inputs to be in.
    n_numerals: int, optional
        How many numerical arguments does the C function expect.
    takes_bool: bool, optional
        Whether the C function takes a bool value as its last argument.

    Returns
    -------
    argtypes:
        A list containing all the arguments' types
    """

    argtypes = []
    for _ in range(npointers):
        argtypes.append(ops.ctypeslib.ndpointer(dtype=dtype, ndim=ndim, shape=shape, flags=flags))
    for _ in range(n_numerals):
        argtypes.append(numeral_type)
    if takes_bool:
        argtypes.append(ctypes.c_bool)
    return argtypes


def make_callable(function: Callable, argtypes: list, restype: Optional[Union[list, Any]] = None) -> None:
    """Supplies the C function with the parameters data types and return type to make it callable from python and
    enable type checking."""
    function.argtypes = argtypes
    function.restype = restype


operating_system = platform.system()
if operating_system == 'Windows':
    SHARED_LIB_FILE_EXTENSION = '.dll'
elif operating_system in ('Linux', 'Darwin'):
    SHARED_LIB_FILE_EXTENSION = '.so'
else:
    raise OSError(
        f"Supported OSes are Windows, Linux and MacOS. Got {operating_system}")

SHARED_LIB_PATH = os.path.join(os.path.dirname(__file__), 'lib', 'c_layers' + SHARED_LIB_FILE_EXTENSION)
if not os.path.exists(SHARED_LIB_PATH):
    raise FileNotFoundError(
        "The compiled shared/dynamic library doesn't exist, you are probably running source distribution code. "
        "To build the package run `python -m build` and install the wheel using `pip install "
        "dist/xrnn-REST-OF-THE-NAME.whl`. If you've already done this, you might be running your commands from the"
        "source distribution directory, and Python will import that code instead of the installed one.")

# Use windll.LoadLibrary on Windows instead of CDLL because it uses stdcall.
# That's useful when passing too many arguments to a C function because it raises a TypeError, making it harder to
# pass the wrong number of arguments to the C functions. Using CDLL on Windows would allow passing too many arguments.
# However, that's not the case for Linux and macOS.
if operating_system == 'Windows':
    functions_cdll = ctypes.windll.LoadLibrary(SHARED_LIB_PATH)
else:
    functions_cdll = ctypes.CDLL(SHARED_LIB_PATH)

# Python caches the imported module so the following module level code will only be executed once, therefor the shared
# library is only loaded once.

# Convolution forward operation that takes float inputs.
convForwardF = functions_cdll.convForwardF
make_callable(convForwardF, make_argtypes_list('float32', npointers=4, n_numerals=11))
# Convolution forward operation that takes double inputs.
convForwardD = functions_cdll.convForwardD
make_callable(convForwardD, make_argtypes_list('float64', npointers=4, n_numerals=11))

convBackwardF = functions_cdll.convBackwardF
make_callable(convBackwardF, make_argtypes_list('float32', npointers=5, n_numerals=11))
convBackwardD = functions_cdll.convBackwardD
make_callable(convBackwardD, make_argtypes_list('float64', npointers=5, n_numerals=11))

maxPoolForwardF = functions_cdll.maxPoolForwardF
make_callable(maxPoolForwardF, make_argtypes_list('float32', npointers=3, n_numerals=10))
maxPoolForwardD = functions_cdll.maxPoolForwardD
make_callable(maxPoolForwardD, make_argtypes_list('float64', npointers=3, n_numerals=10))

maxPoolBackwardF = functions_cdll.maxPoolBackwardF
make_callable(maxPoolBackwardF, make_argtypes_list('float32', npointers=3, n_numerals=10))
maxPoolBackwardD = functions_cdll.maxPoolBackwardD
make_callable(maxPoolBackwardD, make_argtypes_list('float64', npointers=3, n_numerals=10))

avgPoolForwardF = functions_cdll.avgPoolForwardF
make_callable(avgPoolForwardF, make_argtypes_list('float32', npointers=2, n_numerals=10))
avgPoolForwardD = functions_cdll.avgPoolForwardD
make_callable(avgPoolForwardD, make_argtypes_list('float64', npointers=2, n_numerals=10))

avgPoolBackwardF = functions_cdll.avgPoolBackwardF
make_callable(avgPoolBackwardF, make_argtypes_list('float32', npointers=2, n_numerals=10))
avgPoolBackwardD = functions_cdll.avgPoolBackwardD
make_callable(avgPoolBackwardD, make_argtypes_list('float64', npointers=2, n_numerals=10))



"""
* xrnn is a fast, lightweight Python machine learning framework that provides a simple user interface to build and
  train neural networks similar to Keras.
* xrnn only requires one third library dependency which is NumPy (almost likely you had it already installed).
* For more information, look at `layers.py` and `models.py` docstrings, and for a minimal working example of how to use
  the package to train a CNN on MNIST dataset, see `example.py`.
"""

__version__ = "1.1.1"



import sys

if sys.version_info.minor > 6:
    from contextlib import nullcontext
else:
    from contextlib import suppress as nullcontext
from xrnn import layers
from xrnn import losses
from xrnn import ops
import pytest

BATCH_SIZE = 64


def one_hot_encoded(n_classes):
    integers = ops.random.integers(0, n_classes, (BATCH_SIZE,))
    one_hot = ops.zeros((BATCH_SIZE, n_classes))
    one_hot[ops.arange(len(integers)), integers] = 1
    return one_hot


class TestLoss:

    @pytest.mark.parametrize(
        "weight_l2, bias_l2",
        [
            (0.2, 0),
            (0, 0.2),
            (0.15, 0.22),
            (0, 0),
        ]
    )
    def test_regularization_loss(self, weight_l2, bias_l2):
        loss = losses.Loss()
        layer = layers.Dense(100, weight_l2=weight_l2, bias_l2=bias_l2)
        layer.build((BATCH_SIZE, 100))
        loss.trainable_layers.append(layer)
        assert loss.regularization_loss() == layer.weight_l2 * ops.sum(
            ops.square(layer.weights)) + layer.bias_l2 * ops.sum(ops.square(layer.biases))

    def test_calculate(self):
        # This is just to test that the `calculate` method behaves as intended and not for testing if the loss function
        # calculates the loss value correctly, that is tested for each loss function separately.
        loss = losses.BinaryCrossentropy()
        y_true = ops.random.integers(0, 2, (BATCH_SIZE, 1))
        assert loss.calculate(y_true, y_true) == pytest.approx(0, 1.1e-7, 1.1e-7)


class TestCategoricalCrossentropy:

    @pytest.mark.parametrize(
        "y_true, raises",
        [
            (ops.random.integers(0, 10, (BATCH_SIZE, 1)), nullcontext()),
            (ops.random.integers(0, 30, (BATCH_SIZE, 1)), pytest.raises(IndexError)),
            (ops.random.integers(0, 10, (BATCH_SIZE,)), nullcontext()),
            (ops.random.integers(0, 30, (BATCH_SIZE,)), pytest.raises(IndexError)),
            (one_hot_encoded(10), nullcontext()),
            (one_hot_encoded(30), pytest.raises(IndexError))
        ]
    )
    def test_forward(self, y_true, raises):
        with raises:
            loss = losses.CategoricalCrossentropy()
            # The following code is to simulate the output of a dense/softmax layer with 10 neurons.
            y_pred = ops.random.uniform(0, 1, (BATCH_SIZE, 10))
            if y_true.ndim == 2:
                if y_true.shape[1] != 1:  # When labels are actually one-hot encoded.
                    y_true_clipped = ops.argmax(y_true, 1)
                else:
                    y_true_clipped = ops.squeeze(ops.squeeze(y_true.copy()))
            else:
                y_true_clipped = y_true.copy()
            # We don't want to raise IndexError from the test itself, so this avoids that when n_classes > n_neuron.
            y_true_clipped[y_true_clipped >= 10] = 9
            y_pred[ops.arange(len(y_true)), y_true_clipped] = 1  # To match y_pred to y_true to get zero loss.
            assert loss.calculate(y_true, y_pred) == pytest.approx(0, 1e-7, 1.2e-7)  # y_pred values that are equal to
            # zero or one are clipped to config.EPSILON (default is 1e-7) so even if y_true and y_pred exactly match
            # loss isn't going to be exactly zero.

    def test_backward(self):
        loss = losses.CategoricalCrossentropy()
        y_true = ops.random.integers(0, 10, (BATCH_SIZE,))
        y_pred = one_hot_encoded(10)
        one_hot_y_true = ops.eye(len(y_pred[0]))[y_true]
        expected = -one_hot_y_true / (y_pred + 1e-7) / len(y_pred)
        assert ops.allclose(loss.backward(y_true, y_pred), expected)
        assert ops.allclose(loss.backward(ops.expand_dims(y_true, 1), y_pred), expected)
        assert ops.allclose(loss.backward(one_hot_y_true, y_pred), expected)


class TestBinaryCrossentropy:

    @pytest.mark.parametrize(
        "y_true",
        [
            (ops.random.integers(0, 2, (BATCH_SIZE, 1))),
            (ops.random.integers(0, 2, (BATCH_SIZE,))),
            (one_hot_encoded(1)),
            (one_hot_encoded(2)),
            (one_hot_encoded(3)),
        ]
    )
    def test_forward(self, y_true):
        loss = losses.BinaryCrossentropy()
        y_pred = ops.expand_dims(y_true, 1) if y_true.ndim == 1 else y_true
        # y_pred is always going to be 2-dimensional coming out of the network so no need to do this in a real scenario.
        assert loss.calculate(y_true, y_pred) == pytest.approx(0, 1e-7, 1.2e-7)


class TestMSE:

    @pytest.mark.parametrize(
        "y_true",
        [
            (ops.random.random((BATCH_SIZE, 1))),
            (ops.random.random((BATCH_SIZE,))),
            (ops.random.uniform(-1, 1, (BATCH_SIZE, 16, 16, 3))),
            (ops.random.standard_normal((BATCH_SIZE, 2))),
            (ops.random.integers(0, 2768, (BATCH_SIZE, 100))),
        ]
    )
    def test_forward(self, y_true):
        loss = losses.MSE()
        y_pred = ops.expand_dims(y_true, 1) if y_true.ndim == 1 else y_true
        assert loss.calculate(y_true, y_pred) == pytest.approx(0, 1e-7, 1.1e-7)



"""Contains DataHandler class which is used for dealing with user data. See `~DataHandler` docs for more info."""
import sys
from typing import Tuple, Union, Optional

from xrnn import config
from xrnn import ops

if sys.version_info.minor < 8:
    from typing_extensions import Protocol
else:
    from typing import Protocol  # Protocol was added to typing in Python 3.8

ArraysTypeHint = Union[list, ops.ndarray]


class SupportsGetitem(Protocol):
    """Class That is used for annotating parameters that implement __getitem__"""

    def __getitem__(self: 'SupportsGetitem', item: int) -> Tuple[ops.ndarray, ops.ndarray]:
        ...


class DataHandler:
    # The following variables are not treated like class attributes, and they are treated as instance attributes as
    # intended because no value is set to them and only type hints. See 'https://stackoverflow.com/a/69115321'.
    x: Union[SupportsGetitem, list, ops.ndarray]
    y: Optional[Union[list, ops.ndarray]]
    batch_size: int
    shuffle: bool

    def __init__(
            self,
            x: Union[SupportsGetitem, list, ops.ndarray],
            y: Optional[Union[list, ops.ndarray]] = None,
            batch_size: Optional[int] = 32,
            shuffle: Optional[bool] = True) -> None:
        """
        Envelopes the data provided and creates a DataHandler class around them which slices the into batches of size
        `batch_size` and optionally shuffles it, so it can be fed into the neural network for training or inference.

        Parameters
        ----------
        x: list, numpy array, or a generator
            The input features. In case a generator was provided, it has to define __len__ which should return the
            number of batches in the dataset, and __getitem__ which should return a tuple of two numpy arrays, the first
            one is x and the second one is y.
        y: list or numpy array, optional
            Labels. The first axis should be the number of samples in the dataset. Should be None if `x` is a generator.
        batch_size: int
            How many samples should each slice of the data have. If it's set to None, the whole dataset is returned per
            iteration.
        shuffle: bool
            Whether to shuffle the data.

        Raises
        ------
        AttributeError
            If x is a generator that doesn't implement __getitem__ and __len__
        ValueError
            If the data isn't homogenous, if the number of samples in `x` doesn't equal `y`, if `x` is a generator,
            and it returned only one or more than 2 arrays when called, if `validation_split` is too small or too big.
        TypeError
            If `y` is not provided and `x` isn't a generator, if `x` is a generator, and it didn't return numpy arrays,
            if it returned None or one array, if the `x` or `y` couldn't be cast to 'float32' or 'float64'.
        RuntimeError
            If `x` is a generator and an exception was raised when trying to fetch data from it.
        IndexError
            If `x` is a generator, and it returned an empty batch_x or batch_y.

        Notes
        -----
        * The word "generator" here is used to denote a class that implements __getitem__ and __len__ and not the
          conventional meaning of it in Python. Used it for a lack of a better word.
        * If both x and y are numpy arrays and have a different data type than the default data type used within this
          package (float32, which can be changed), they will be copied and cast to it increasing memory usage. To avoid
          this, cast them to the default data type before making any operation on them (like `model.train`) or change
          the default data type to match the data by calling
          `from xrnn import config; config.set_default_dtype(x.dtype)`

        Examples
        --------
        >>> samples = ops.random.random((2000, 128))  # 1000 random samples where each one has 128 features.
        >>> labels = ops.random.integers(0, 10, (2000, 1))  # 10 classes.
        >>> dataset = DataHandler(samples, labels, 32, True)  # batch size of 32, and shuffle set to True.
        >>> batch = dataset[0]  # Retrieve the batch at index 0
        >>> len(batch)
        2
        >>> x_batch, y_batch = batch
        >>> len(x_batch)
        32
        >>> len(dataset)  # Number of batches in the dataset.
        63
        """
        self.batch_size = batch_size
        self.shuffle = shuffle
        self._dtype = config.DTYPE
        self.check_x, self.check_y = None, None  # Used as a placeholder for the data received from the generator when
        # checking its validity. Save the data because the generator may keep track of the data it returned and calling
        # the generator to get the data at index 0 twice (when actually iterating over it) may break its logic. So we
        # save them and return them only the first time we iterate over it.
        if y is None:  # A generator/iterator is (should've been) provided.
            if isinstance(x, (list, tuple, ops.ndarray)):
                raise TypeError(
                    f"Only `x` was provided, and it's of type: {type(x)} not a generator/iterator object. This might "
                    f"indicate that both x and y are in the same array, for e.g. [x_train, y_train] is passed to the "
                    f"`x` argument, if this is the case, please separate them and pass them as "
                    f"individual arrays to both `x` and `y` arguments respectively, or pass a generator that complies"
                    f"with the documentation of the `x` parameter for generators.")
            if self.validate_generator(x):
                self.x, self.y = x, y
        else:
            if len(x) != len(y):
                raise ValueError(
                    f"The number of samples in both the input features and the labels must be the same. "
                    f"Got x: {len(x)} samples, y: {len(y)} samples.")
            self.x = self.to_ndarray(x)
            self.y = self.to_ndarray(y)
            self.batch_indices = list(range(len(self)))  # Used in case shuffle was set to True to access random
            # indices of the array instead of copying/creating a new randomized array.
        config.CREATED_OBJECTS.append(self)

    @property
    def dtype(self) -> config.DtypeHint:
        """Returns the data type of the data."""
        return self._dtype

    @dtype.setter
    def dtype(self, new_dtype: config.DtypeHint) -> None:
        new_dtype = config.parse_datatype(new_dtype)
        self._dtype = new_dtype
        if self.y is not None:
            self.x = self.to_ndarray(self.x)
            self.y = self.to_ndarray(self.y)

    def validate_generator(self, generator: SupportsGetitem) -> bool:
        """Validates if the generator can be used."""
        if not hasattr(generator, '__len__'):
            raise AttributeError(
                "The provided generator must define `__len__` magic method that should return the number of batches"
                " in the dataset (n_samples // batch_size).")
        if not hasattr(generator, '__getitem__'):
            raise AttributeError(
                "The provided generator must define `__getitem__` magic method that returns the batches of data.")
        try:
            self.check_x, self.check_y = generator[0]
            if not isinstance(self.check_x, ops.ndarray) or not isinstance(self.check_y, ops.ndarray):
                raise TypeError(
                    f"The values returned from the generator must be numpy arrays. "
                    f"Got x: {type(self.check_x)}, y: {type(self.check_y)}.")
            self.check_x, self.check_y = self.to_ndarray(self.check_x), self.to_ndarray(self.check_y)
            if len(self.check_x) != len(self.check_y):
                raise ValueError(
                    f"The number of samples in both the input features and the labels must be the same. "
                    f"Got x: {len(self.check_x)} samples, y: {len(self.check_y)} samples.")
        except ValueError as e:
            if str(e) == 'not enough values to unpack (expected 2, got 1)':
                # Chaining the thrown exception with ours to be more verbose and in case the exception was actually
                # thrown from within the generator and not because of the unpacking of the returned values.
                raise ValueError(
                    'The generator returned only one value. It should return two arrays instead. This probably '
                    'happened because the generator returned an iterable (for e.g. a numpy array) containing only one '
                    'value.') from e
            if str(e) == 'too many values to unpack (expected 2)':
                raise ValueError(
                    "The generator returned more than two values. It should only return batch_x, batch_y.") from e
            raise e
        except TypeError as e:
            if 'cannot unpack non-iterable' in str(e):
                raise TypeError("The generator only returned one value or didn't return anything at all. "
                                "It should return two values, batch_x and batch_y.") from e
            raise e  # Raise the exception thrown if it doesn't match the expected exception
        except Exception as e:
            raise RuntimeError(
                "The previous error occurred while trying to retrieve the first item from the generator.") from e
        return True

    def to_ndarray(self, arr: ops.ndarray) -> ops.ndarray:
        """Converts the data to a numpy array if it's not already a numpy array, or converts to dtype if it has a
        different dtype. Raises a ValueError if the data couldn't be converted."""
        try:
            if not isinstance(arr, ops.ndarray):
                return ops.array(arr, dtype=self.dtype)
            if str(arr.dtype) != self.dtype:
                return arr.astype(self.dtype, 'C')
            return arr
        except ValueError:
            raise ValueError("The data must be homogenous, this means that all the elements have the same shape, "
                             "and that data provided doesn't satisfy this requirement.")
        except TypeError as e:
            raise TypeError(
                f"Could not convert the input array which has a type of {type(arr)} to a numpy array with a "
                f"{self.dtype} datatype.") from e

    def __getitem__(self, idx: int) -> Tuple[ops.ndarray, ops.ndarray]:
        """Returns batch_x, batch_y each containing `batch_size` number of samples at index `idx`."""
        if self.y is None:
            if self.check_x is not None:
                batch_x, batch_y = self.check_x.copy(), self.check_y.copy()
                self.check_x, self.check_y = None, None
                return self.to_ndarray(batch_x), self.to_ndarray(batch_y)  # Just to convert the dtype if necessary.
            try:
                batch_x, batch_y = self.x[idx]
                if not len(batch_x) or not len(batch_y):
                    raise IndexError(
                        "A an empty array was returned. This is probably due to out of bounds indexing. A good reason "
                        "why that might've happened is that the `__len__` method of generator is returning an "
                        "incorrect number of batches which leads to accessing slices of the array that are empty.")
                return self.to_ndarray(batch_x), self.to_ndarray(batch_y)  # Just to convert the dtype if necessary.
            except IndexError as e:
                raise e  # Raise our IndexError if it occurred to not catch in proceeding catch.
            except Exception as e:
                # Raise a runtime error from the exception thrown by self.x[idx] because it's the direct cause.
                raise RuntimeError(
                    f"The generator threw the previous exception when it's called to get data at index {idx}.") from e
        if idx == 0 and self.shuffle:  # Shuffle the data everytime we loop over it from the start.
            self.batch_indices = ops.random.permutation(self.batch_indices)
        if not self.batch_size:
            return self.x, self.y
        st_idx, end_idx = self.batch_indices[idx] * self.batch_size, (self.batch_indices[idx] + 1) * self.batch_size
        return self.x[st_idx: end_idx], self.y[st_idx: end_idx]

    def __len__(self) -> int:
        """Returns the number of batches in the dataset"""
        if self.y is None:
            return int(len(self.x))
        if not self.batch_size:
            return 1
        steps = len(self.x) // self.batch_size
        if steps * self.batch_size < len(self.x):
            steps += 1
        return steps

    @staticmethod
    def train_test_split(
            x: ArraysTypeHint, y: ArraysTypeHint, validation_split: float
    ) -> Tuple[ArraysTypeHint, ArraysTypeHint, ArraysTypeHint, ArraysTypeHint]:
        """
        A function that splits the dataset to `1 - validation_data` % training data and
        `validation_split` % validation data.

        Parameters
        ----------
        x: list or numpy array
            Training features.
        y: list or numpy array
            labels.
        validation_split: float between (0, 1)
            The percentage of the dataset to use as validation data.

        Returns
        -------
        split_data: tuple of four arrays
            x_train, y_train, x_test, y_test.

        Raises
        ------
        ValueError
            If validation_split is not between 0 and 1, if validation_split is too small or too big.
        """
        train_samples = round(len(x) * (1 - validation_split))
        err = 'The resulting number of training samples is equal to {}, this means that `validation_split` value is ' \
              'too {}, please make it {} or add more sample.'
        if train_samples >= len(x):
            raise ValueError(err.format('len(x)', 'small', 'bigger'))
        if train_samples <= 0:
            raise ValueError(err.format('0', 'big', 'smaller'))
        return x[:train_samples], y[:train_samples], x[train_samples:], y[train_samples:]



"""Defines the `Model` class that is used for building, training, and using neural networks. See `Model`'s documentation
for more info and example usage."""
import time
import warnings
from collections import deque  # Used to implement moving average used to calculate the average step time.
from functools import reduce
from typing import Union, List, IO
import json
import os
import zipfile
import tempfile
import io
import sys

from xrnn import activations
from xrnn import config
from xrnn import layer_utils
from xrnn import layers
from xrnn import losses
from xrnn import metrics
from xrnn import ops
from xrnn import optimizers
from xrnn.data_handler import DataHandler, SupportsGetitem


class Model:

    def __init__(self) -> None:
        """
        Groups a linear stack of layers. The input flows linearly through each layer.
        Some useful methods provided by this class. Please take a look at each method's documentation for more
        information about what the method does and how to use it.
         * Layers can be added via `model.add`.
         * Set a loss function and an optimization algorithm by calling `model.set`.
         * Training the neural network/model can be done by calling `model.train`.
         * Validating the network can be done by passing validation data or validation split to `model.train` or by
           calling `model.validate(x_test, y_test)`
         * Using the model for inference/prediction can be done by calling `model.predict` or `model.inference`.
         * Printing a summary of the model's layers showing various useful information can be done by calling
           `model.summary`.
         * Getting the memory consumption of the model in bytes can be done by calling `model.mem_usage`.
         * `model.save` saves the whole model to disk using pickle.
         * `model.save_parameters` saves the model's parameters to disk.

        Notes
        -----
        * This method of constructing a neural network doesn't allow for having multiple inputs to the network or to a
          layer unlike `keras.models.Model` for example. This is more like `keras.models.Sequential`.
        * The model is lazily built, meaning that layers' weights and biases aren't initialized/created until explicitly
          calling `model.build(input_shape)` or implicitly when an input is passed to the model like when calling
          `predict` for example. Until that happens, the model is just a blueprint for constructing the neural network,
          therefor, creating instances of `Model` are very cheap.
        """

        self.layers = []
        self.trainable_layers = []
        self.loss = None
        self.optimizer = None
        self.built = False
        self.input_shape = None
        self.batch_size = None
        self._output_shape = None

    def set(
            self,
            optimizer: Union[optimizers.Optimizer, config.Literal['adam', 'sgd', 'rmsprop', 'adagrad']],
            loss: Union[losses.Loss, config.Literal['mse', 'categorical_crossentropy', 'binary_crossentropy']]) -> None:
        """
        Sets the model loss and optimizer to use.

        Parameters
        ----------
        optimizer: Optimizer, str
            An optimizer instance to use for optimizer the model. See `xrnn.optimizers` for available ones or a string.
            In the case of a string, it must be one of the following: adam, sgd, rmsprop, adagrad, and an optimizer of
            that type is going to be created with its default parameters.
        loss: Loss, str
            A loss object instance used to calculate the model's loss. See `xrnn.losses` for available ones or a string.
            In case of a string, it must be one of the following: mse, categorical_crossentropy, binary_crossentropy.
        """
        if isinstance(optimizer, str):
            opt_str = {
                'adam': optimizers.Adam, 'sgd': optimizers.SGD,
                'rmsprop': optimizers.RMSprop, 'adagrad': optimizers.Adagrad}
            if optimizer.lower() not in opt_str:
                raise ValueError(f"Invalid optimizer. Must be one of {opt_str.keys()}. Got {optimizer}.")
            optimizer = opt_str[optimizer.lower()]()
        if isinstance(loss, str):
            loss_str = {
                'mse': losses.MSE,
                'categorical_crossentropy': losses.CategoricalCrossentropy,
                'binary_crossentropy': losses.BinaryCrossentropy}
            if loss.lower() not in loss_str:
                raise ValueError(f"Invalid loss. Must be one of {loss_str.keys()}. Got {loss}")
            loss = loss_str[loss.lower()]()
        self.optimizer = optimizer
        self.loss = loss
        if self.loss and self.trainable_layers:
            self.loss.trainable_layers = self.trainable_layers

    def add(self, layer) -> None:
        """Add layers to the network in sequential order."""
        if hasattr(layer, 'weights'):
            self.trainable_layers.append(layer)
            if self.loss:  # If the loss is set first, then layers are added.
                if layer not in self.loss.trainable_layers:
                    self.loss.trainable_layers.append(layer)
        if self.built:
            if layer.built is None:
                layer.input_shape = self.output_shape
            elif layer.built is False:
                layer.build(self.output_shape)
            else:
                if layer.input_shape != self.output_shape:
                    raise ValueError(
                        f"Layer {layer.name} expects inputs with shape {layer.input_shape}. Got input_shape="
                        f"{self.output_shape}.")
            self._output_shape = layer.compute_output_shape(self.output_shape)
        self.layers.append(layer)

    @property
    def output_shape(self):
        """Returns the output shape of the model."""
        if not self.built:
            raise ValueError(
                'The model has not been built, please build the model first by calling `model.build(input_shape)`')
        return self._output_shape

    def build(self, input_shape: tuple) -> None:
        """
        Build the neural network (initialises all the layers parameters) using the specified input shape.

        Notes
        -----
         - That the first value in `input_shape` should be batch size not the number of samples in the dataset.
         - Layers can only be built once, so calling `build()` a second time would raise a ValueError.
        """
        input_shape = tuple(input_shape)
        self.batch_size = input_shape[0]
        self.built = True
        self.input_shape = input_shape
        for i in range(len(self.layers)):
            layer = self.layers[i]
            built = getattr(layer, 'built', None)
            if built is False:
                activation = None
                if i + 1 < len(self.layers) and isinstance(self.layers[i + 1], activations.ReLU):
                    activation = 'relu'
                layer.build(input_shape, activation)
            elif built is True:
                if input_shape != layer.input_shape:
                    raise ValueError(
                        f"Layer {layer.name} expects inputs with shape {layer.input_shape}. Got input_shape="
                        f"{input_shape}.")
            else:
                expected_dims = getattr(layer, 'expected_dims', None)
                if expected_dims and len(input_shape) != expected_dims:
                    raise ValueError(
                        f"Layer {layer.name} expects inputs to be {expected_dims}D. Got {input_shape}.")
                layer.input_shape = input_shape
            input_shape = layer.compute_output_shape(input_shape)
        self._output_shape = input_shape

    def forward(self, inputs: ops.ndarray, training: bool = True) -> ops.ndarray:
        """
        Passes `inputs` through the whole network.

        Parameters
        ----------
        inputs: array_like
            The input to the pass through the network.
        training: bool
            Whether to set the layers to training mode. The behaviour of some layers differs between training and
            inference modes, so set this parameter accordingly.

        Returns
        -------
        output: array_like
            The output of the network.
        """
        if not self.built:
            self.build(inputs.shape)
        before_states = [layer.training for layer in self.layers]  # Save the training state of the layer before
        # changing it, so it can be reset after changing the states for all layers. This is needed because the user
        # might set a layer training state to something other than the default, so we need to keep track of that.
        for layer in self.layers:
            layer.training = training
        # Applies the forward method of each layer in a chain (last_layer.forward(second_to_last.forward(...)) and so on
        output = reduce(lambda x, layer_forward: layer_forward(x), [layer.__call__ for layer in self.layers], inputs)
        for layer, state in zip(self.layers, before_states):
            layer.training = state
        return output

    def __call__(self, inputs: ops.ndarray, training: bool = True) -> ops.ndarray:
        """Calls the model's `forward` method, passing `inputs` through the network."""
        return self.forward(inputs)

    def backward(self, y_true: ops.ndarray, y_pred: ops.ndarray) -> ops.ndarray:
        """Backpropagate through the whole network."""
        # Fast path for softmax + categorical_crossentropy classification.
        if isinstance(self.loss, losses.CategoricalCrossentropy) and isinstance(self.layers[-1], activations.Softmax):
            # When using a softmax - categorical_crossentropy classifier, the derivative becomes much simpler when
            # solved w.r.t both of them, it becomes subtracting 1 from the predicted confidence at the correct index,
            # thus speeding up the calculation a lot (~7x faster).
            if len(y_true.shape) == 2:
                y_true = ops.argmax(y_true, axis=1)
            d_values = y_pred.copy()
            d_values[ops.arange(len(y_true)), y_true.astype('int')] -= 1
            d_inputs = d_values / len(y_true)
            layers_to_backward = self.layers[-2::-1]  # The whole list reversed, excluding the last element.
        else:
            d_inputs = self.loss.backward(y_true, y_pred)
            layers_to_backward = self.layers[::-1]
        return reduce(
            lambda x, layer_backward: layer_backward(x), [layer.backward for layer in layers_to_backward], d_inputs)

    @staticmethod
    def update_progressbar(info: dict, size: int = 30) -> None:
        """Prints a progressbar on one line and keeps updating it."""
        curr_step, tot_steps = info['step'], info['steps']
        print(f"\rstep: {curr_step:{len(str(tot_steps))}d}/{tot_steps} ", end='')  # first part: "step/total_steps".
        progress = int(size * curr_step / tot_steps)
        left = int(size * (tot_steps - curr_step) / tot_steps)
        while progress + left < size:
            left += 1
        print(f"[{'=' * progress + '.' * left:{size}s}] - ", end='')  # the progress bar part.
        print(f"{curr_step / tot_steps:.0%} - ", end='')  # the percentage part.
        epoch_time = info.get('epoch_time')
        if epoch_time:
            print(f"Took: {layer_utils.time_unit_converter(epoch_time)} - ", end='')
        else:
            print(f"ETA: {layer_utils.time_unit_converter(info['avg_step_time'] * (tot_steps - curr_step))} - ", end='')
        for i, key in enumerate(list(info.keys())[4:]):  # other info like val acc/loss. Starting from the 4th element
            # because the first 4 elements are already printed above before the for loop.
            print(f"{key}: {info[key]:.3f}", end=' - ' if i < len(info) - 5 else '')
        if epoch_time:  # If it's not None, it means the epoch ended.
            print()  # Add a new line after the epoch ends to not print next to the progressbar because it doesn't
            # end with a newline.

    def train(
            self,
            x: Union[ops.ndarray, list, SupportsGetitem],
            y: Union[ops.ndarray, list] = None,
            batch_size: int = 32,
            epochs: int = 1,
            shuffle: bool = True,
            validation_data: Union[List[Union[list, ops.ndarray]], SupportsGetitem] = None,
            validation_split: float = 0.0,
            validation_freq: int = 1,
            steps_per_epoch: int = None,
            print_every: int = 1,
            disable_sleep: bool = False) -> None:
        """
        Fits the model to the data.

        Parameters
        ----------
        x:
            Input features (data) to train on. This can be a list, a NumPy array, or a custom object. In the case
            of the custom object, it has to meat a few guidelines. It must define both `__getitem__` and `__len__`
            magic attributes, where `len(obj)` returns the number of batches in the datatest and `obj[batch_idx]`
            returns a tuple containing batch x and batch y. If this custom object is passed, `y` has to be None.
        y: list, numpy array, optional
            Input labels. This can be a list or a NumPy array. Labels should be 2-dimensional when the loss is
            binary crossentropy, mean-squared error or categorical crossentropy, and the labels are one-hot encoded.
        batch_size: int, optional
            How many samples to pass through the whole network at once. Batch size has a direct impact on
            memory consumption and performance, so you want to find a value that utilizes the cpu/gpu without running
            out of memory. Ignored if `x` is a `DataHandler` like object.
        epochs: int, optional
            How many times to go through (train) the whole dataset.
        shuffle: bool, optional
            Whether to randomly shuffle the dataset. Ignored if `x` is `DataHandler` like object.
        validation_data: list, numpy array, DataHandler, optional
            A list containing [x_test, y_test] to validate the model. During the model validation,
            the model's weights and biases aren't updated. It can be the same type of objects as `x`.
        validation_split: float, optional
            The fraction of the training data (x and y) to use for validation. Not supported when `x` is `DataHandler`
            object. If both `validation_split` and `validation_data` are provided, `validation_data` is used.
        validation_freq: int, optional
            Every how many epochs to validate the model. Default is at the end of each epoch. If set
            to zero or None, no validation is performed.
        steps_per_epoch: int, optional
            How many steps (batches) to train for per epoch. Can useful for rapidly testing
            different model hyperparameters when the dataset is huge and there's no need to train for a full epoch to
            get meaningful results.
        print_every: int, optional
            Every how many epochs to print the model progress (in progressbar form). Can be set to zero,
            None, or False to disable printing anything (train silently).
        disable_sleep: bool optional
            Whether to allow the computer to go to sleep during training. This only works on windows.
        """
        if self.optimizer is None or self.loss is None:
            raise ValueError("The optimizer and loss aren't set yet, please call `model.set(opt, loss)`")
        if validation_data is not None:
            if not isinstance(validation_data, (list, tuple)):
                validation_data = [validation_data, None]
        elif validation_split:
            if y is None:
                raise TypeError("`validation_split` is not supported when `x` is a generator like/DataHandler object.")
            x, y, x_test, y_test = DataHandler.train_test_split(x, y, validation_split)
            validation_data = [x_test, y_test]
        val_data_handler = DataHandler(*validation_data) if validation_data else None
        config.set_sleep_state(not disable_sleep)
        start_t = time.time()
        print_every = print_every or epochs + 1  # This way if it's set to zero, None or False updates are never printed
        data_handler = DataHandler(x, y, batch_size, shuffle)
        steps_per_epoch = min(steps_per_epoch, len(data_handler)) if steps_per_epoch else len(data_handler)
        self.batch_size = data_handler.batch_size
        acc = metrics.Accuracy(self.loss)
        # Keep the most recent 10% number of the total steps.
        recent_steps_time = deque(maxlen=int(len(data_handler) * 0.1) or 1)
        for epoch in range(epochs):
            epoch_start_t = time.perf_counter()
            epoch_loss = 0
            epoch_acc = 0
            progress_info = {
                'step': 0, 'steps': steps_per_epoch,
                'avg_step_time': 0, 'epoch_time': None, 'loss': epoch_loss, 'acc': epoch_acc}
            self.loss.reset_count()
            acc.reset_count()
            should_print = (epoch + 1) % print_every == 0
            if should_print:
                print(f'{"" if epoch == 0 else chr(10)}Epoch {epoch + 1}/{epochs}')
            # Add a \n character before the epoch number after epoch one to not print at the end of the progressbar.
            # chr(10) is the unicode presentation of a new line {\n} character, and it's used here because backslash
            # characters can't be used inside an f-string lateral. I know there must be a more readable way of
            # accomplishing this goal like using a blank `print()` after the for loop, but I like this one more.
            for batch_idx in range(len(data_handler)):
                step_start_t = time.perf_counter()
                batch_x, batch_y = data_handler[batch_idx]
                output = self.forward(batch_x, training=True)
                epoch_loss = self.loss.calculate(batch_y, output)
                epoch_acc = acc.calculate(batch_y, output)
                self.backward(batch_y, output)
                for layer in self.trainable_layers:
                    if layer.training:  # Only update parameters for layers that have training set to True. This check
                        # is performed every time because the layer might be trained for a certain number of steps then
                        # frozen, the same reason why all `trainable` layers are added to a list and not just the
                        # `trainable`s.
                        self.optimizer.update_params(layer)
                self.optimizer.iterations += 1
                recent_steps_time.append(time.perf_counter() - step_start_t)
                cur_len = len(recent_steps_time)
                avg_step_time = (sum(
                    [(cur_len - n) * n_step_time for n, n_step_time in enumerate(recent_steps_time)])
                                 / (0.5 * (cur_len * (cur_len + 1))))  # Weighted moving average.
                progress_info.update(
                    {'step': batch_idx + 1, 'avg_step_time': avg_step_time, 'loss': epoch_loss, 'acc': epoch_acc})
                if should_print:
                    self.update_progressbar(progress_info)
                if batch_idx + 1 == steps_per_epoch:
                    break
            self.optimizer.epoch += 1
            if val_data_handler and validation_freq and (epoch + 1) % ops.lcm(validation_freq, print_every) == 0:
                # Only print if epoch is multiple of lcm(validation_freq, print_every). For e.g., if print_every = 3 and
                # validation_freq = 4, validate every 12th epoch.
                val_res = self.evaluate(val_data_handler)
                progress_info.update({'val_loss': val_res['loss'], 'val_acc': val_res['acc']})
            if should_print:
                progress_info.update({'epoch_time': time.perf_counter() - epoch_start_t})
                self.update_progressbar(progress_info)
        # Because the progressbar prints how long an epoch took, so there's no reason to print it again if epochs=1.
        if epochs > 1 and print_every <= epochs:
            print(f"Training took {(time.time() - start_t):.3f} seconds")
        config.set_sleep_state(True)  # Resets the computer sleep state. Doesn't matter if the sleep state was changed
        # or not before, because this function works based on the thread it's called from not system-wide.

    def evaluate(
            self,
            x: Union[ops.ndarray, list, SupportsGetitem],
            y: Union[ops.ndarray, list] = None,
            batch_size: int = 32) -> dict:
        """Evaluates/validates the model using `x` and `y` data."""
        if not isinstance(x, DataHandler):
            data_handler = DataHandler(x, y, batch_size, shuffle=False)
        else:
            data_handler = x
        val_loss = 0
        val_acc = 0
        acc = metrics.Accuracy(self.loss)
        self.loss.reset_count()
        for batch_idx in range(len(data_handler)):
            batch_x, batch_y = data_handler[batch_idx]
            output = self.forward(batch_x, training=False)
            val_loss = self.loss.calculate(batch_y, output)
            val_acc = acc.calculate(batch_y, output)
        return {'loss': val_loss, 'acc': val_acc}

    def inference(self, x: ops.ndarray, batch_size: int = None) -> ops.ndarray:
        """
        Provides an interface to use model for inference/prediction.

        This method provides some useful features over using the `forward` method directly. First, it sets the training
        flag in the layers to false, which makes performing a forward pass through the network faster and use less
        memory because the layers don't save any intermediate calculations/inputs/outputs they have/do because these
        values are only required for backpropagation which isn't going to be performed here. Second, it allows
        performing the forward pass (inference) in batches.

        Parameters
        ----------
        x: numpy array
            Data to perform inference/predict using.
        batch_size: int
            Number of samples to perform a prediction on each step. Lower the number if encountering oom issues.
            If None, the whole input is passed at once.

        Returns
        -------
        predictions: numpy array
            predicted output that has the same shape as the output layer of the network.

        Notes
        -----
        When batch_size is set to `None`, the whole input `x` is passed at once (default).
        """
        if not batch_size:
            return self.forward(x, training=False)
        data_batches = [x[i * batch_size: (i + 1) * batch_size] for i in range(len(x) // batch_size + 1 or 1)]
        predictions = []
        for batch in data_batches:
            predictions.append(self.forward(batch, training=False))
        return ops.vstack(predictions)

    predict = inference  # Alias

    def mem_usage(self, batch_size: int = None) -> int:
        """
        Returns the total memory used by the model's layers' parameters + gradients + activations. **Note** this is
        not equivalent to calling `getsizeof` recursively on the model instance because this method doesn't calculate
        the memory used by the Python objects themselves, this is because the size of bare objects is minuscule compared
        to the arrays holding numbers for the different layers' attributes even for a small network.

        Parameters
        ----------
        batch_size: int, optional
            How many samples each slice of the data contains. This number largely determines the memory consumption,
            because some inputs/outputs of the layer are saved and the more samples a batch has, the more memory is
            consumed. Defaults to None, which means use the batch size the model was built with/trained on. You can
            provide any number for this parameter to see how much memory is going to be consumed by the network without
            actually training it.

        Returns
        -------
        tot_mem: int
            Memory consumed by the model in bytes.
        """
        if not self.built:
            raise ValueError("The model must be built before calling this method. Please call `model.build()` first.")
        batch_size = batch_size or self.batch_size  # Give priority to the provided batch_size when picking.
        input_shape = (batch_size,) + self.input_shape[1:]
        tot = 0
        for layer in self.layers:
            tot += layer_utils.layer_memory_consumption(
                layer, input_shape, True, isinstance(self.optimizer, optimizers.Adam))[-1]
            input_shape = layer.compute_output_shape(input_shape)
        return tot

    def summary(self, batch_size: int = None, memory_consumption: bool = True) -> None:
        """
        Prints a summary of the whole model. This summary includes:
         1. Each layer's type, number of parameters, input_shape and output_shape.
         2. Total number of parameters in the model.
         3. Trainable parameters.
         4. Non-trainable parameters.
         5. Total number of layers the model has.
         6. parameters/gradient/activations/total memory consumption.

        Notes
        -----
        * Parameters memory consumption calculates how much memory all the layers weights and biases take.
        * Gradient memory consumption calculates how much memory all the layers gradients (dL/d_weights, dL/d_biases)
          and the optimizer stats (weight momentum cache, biases momentum cache, etc.) take.
        * Activations memory consumption calculates how much memory al the layers saved inputs/output take.
        * For memory consumption, it calculates how
          much the activation/gradients are going to take when a forward/backward pass is performed, so when the model
          is first built and the gradients haven't been calculated yet, it's still going to print (how much) memory
          they are going to take.
        * When no optimizer is set for the model, it's assumed that the model is only going to be used for inference,
          so no activation/gradients are going to be created thus 0 memory consumption.

        Parameters
        ----------
        batch_size: int, optional
            How many samples each slice of the data contains. This number largely determines the memory consumption,
            because some inputs/outputs of the layer are saved and the more samples a batch has, the more memory is
            consumed. Defaults to None, which means use the batch size the model was built with/trained on. You can
            provide any number for this parameter to see how much memory is going to be consumed by the network without
            actually training it.
        memory_consumption: bool, optional
            Whether to print various memory consumption metrics. Default is True.
         """
        if not self.built:
            raise ValueError(
                "Please call `model.build()` before calling `summary()` because the model has to be built.")
        batch_size = batch_size or self.batch_size  # Give priority to the provided batch_size when picking.
        if batch_size > 1024:
            if not batch_size & (batch_size - 1) == 0:  # check if batch size is not a power of 2
                warnings.warn(
                    "Usually `batch_size` is a power of 2 and isn't this large. Are you sure that the first "
                    "value in the input shape tuple that was passed to `model.build()` is the batch size and not the "
                    "number of samples in the dataset? If so please ignore this warning.", UserWarning)
        tot_params = 0
        trainable_params = 0
        non_trainable_params = 0
        tot_mem_params = 0
        tot_mem_grads = 0
        tot_mem_activations = 0
        tot_mem = 0
        column_titles = [['Layer', 'params #', 'input_shape', 'output_shape']]
        input_shape = (batch_size,) + self.input_shape[1:]
        for layer in self.layers:
            params = layer.weights.size + layer.biases.size if hasattr(layer, 'weights') else 0
            bn_non_trainable = layer.weights.size * 2 if isinstance(layer, layers.BatchNormalization) else 0
            non_trainable_params += bn_non_trainable
            if layer.training:
                trainable_params += params
            else:
                non_trainable_params += params
            params += bn_non_trainable
            tot_params += params
            if memory_consumption:
                mem_params, mem_grads, mem_activation, mem_tot = layer_utils.layer_memory_consumption(
                    layer, input_shape, True, isinstance(self.optimizer, optimizers.Adam))
                tot_mem_params += mem_params
                tot_mem_grads += mem_grads
                tot_mem_activations += mem_activation
                tot_mem += mem_tot
            output_shape = layer.compute_output_shape(input_shape)
            column_titles.append([layer.name, params, str(input_shape), str(output_shape)])
            input_shape = output_shape
        layer_utils.print_table(
            column_titles,
            {
                "Trainable params": f'{trainable_params:,}',
                "Non-trainable params": f'{non_trainable_params:,}',
                "Total params": f'{tot_params:,} ({layer_utils.to_readable_unit_converter(tot_mem_params)})',
                "Gradients mem consumption": layer_utils.to_readable_unit_converter(tot_mem_grads),
                "Activations mem consumption": layer_utils.to_readable_unit_converter(tot_mem_activations),
                "Total mem consumption": layer_utils.to_readable_unit_converter(tot_mem)})

    def get_params(self) -> List[List[ops.ndarray]]:
        """Returns a list of lists containing each layer's parameters."""
        if not len(self.trainable_layers):
            raise ValueError('The model does not have any trainable layers.')
        return [layer.get_params() for layer in self.trainable_layers]

    def set_params(self, params: List[List[ops.ndarray]], keras_weights: bool = False) -> None:
        """
        Sets the model's layers' parameters.

        Parameters
        ----------
        params: list of lists
            A list of lists containing each layer's parameters.
        keras_weights: bool, optional
            Whether the parameters come from a keras model. Defaults to False

        Raises
        ------
        IndexError
            If the number of lists containing the parameters doesn't match the number of layers in the model that
            have parameters.
        ValueError
            If the model hasn't been built, if there's a shape mismatch between a layer's parameters and its
            corresponding new parameters.
        """
        if len(params) != len(self.trainable_layers):
            raise IndexError(
                f"Number of supplied parameters does not match number of layers with parameters in the model. Number "
                f"of layers is {len(self.trainable_layers)}, got {len(params)}.")
        for layer, layer_params in zip(self.trainable_layers, params):
            if isinstance(layer, layers.Conv2D):
                layer.set_params(layer_params, keras_weights)
            else:
                layer.set_params(layer_params)

    def save_params(self, path: str) -> None:
        """
        Saves the model's parameters to disk. The model must be built before calling this method.

        See Also
        save: Saves the model's config and optimizer along its parameters.
        """
        model_params = self.get_params()
        model_params_dict = {}
        for layer, layer_params in zip(self.trainable_layers, model_params):
            for i, param in enumerate(layer_params):
                model_params_dict[f"{layer.name}_{i}"] = param
        ops.savez(path, **model_params_dict)

    def load_params(self, path: Union[str, IO]) -> None:
        """Loads the model's parameters from disk. The model must be built before calling this method."""
        model_params = []
        # This is needed for Python 3.6, since the `seek()` method wasn't implemented for `ZipExtFile` until 3.7.
        # https://bugs.python.org/issue22908
        if not isinstance(path, str) and sys.version_info[:2] == (3, 6):
            path = io.BytesIO(path.read())
        with ops.load(path) as model_params_dict:
            layer_params = []
            for i, key in enumerate(model_params_dict):
                if i != 0 and key.endswith('0'):
                    model_params.append(layer_params)
                    layer_params = []
                layer_params.append(model_params_dict[key])
            model_params.append(layer_params)
        self.set_params(model_params)

    def save(self, path: str) -> None:
        """
        Saves the whole model to disk. This method is different to `save_params` in that it saves the model's
        architecture, input shape and optimizer state, meaning you can load the entire model later without the need
        to know the model's architecture and input shape.
        .. note:: The model is saved as a zip file with '.xmodel' file extension.
        """
        # TODO: Save the layers' parameters cache and momentum cache used by the optimizer.
        path += '.xmodel' if not path.endswith('.xmodel') else ''
        with zipfile.ZipFile(path, 'w') as model_file:
            json.dumps(self.get_config()['opt_config'])
            model_file.writestr('model_config.json', json.dumps(self.get_config()))
            model_params_path = os.path.join(tempfile.gettempdir(), 'temp_model_params.npz')
            self.save_params(model_params_path)
            model_file.write(model_params_path, 'model_params.npz')
            os.remove(model_params_path)

    @staticmethod
    def load(path: str) -> 'Model':
        """Loads a model from disk."""
        with zipfile.ZipFile(path) as saved_model:
            with saved_model.open(
                    'model_config.json') as config_file, saved_model.open('model_params.npz') as model_params_file:
                # Fun fact: Python versions <3.9 don't support parenthesized context managers because the parser they
                # use `LL(1)` sucks. See
                # stackoverflow.com/questions/68924790/parenthesized-context-managers-work-in-python-3-9-but-not-3-8
                model_config = json.load(config_file)
                model = Model.from_config(model_config, True)
                model.load_params(model_params_file)
        return model

    def get_config(self) -> dict:
        """
        Returns the model's configuration. It contains the following keys:
         - 'layers_configs': A list that contains each layer's config.
         - 'opt_config': The optimizer configuration if it's set, None otherwise.
         - 'loss': Loss class name (e.g, CategoricalCrossentropy) if its set, None otherwise.

        Returns
        -------
        model_config: dict
            The model's config.
        """
        layers_configs = [layer.get_config() for layer in self.layers]
        opt_config, loss = None, None
        if self.optimizer:
            opt_config = self.optimizer.get_config()
            loss = type(self.loss).__name__
        return {'layers_configs': layers_configs, 'opt_config': opt_config, 'loss': loss}

    @classmethod
    def from_config(cls, model_config: dict, build_if_can: bool = False) -> 'Model':
        """
        Creates a new `Model` instance from the given config.

        Parameters
        ----------
        model_config: dict
            A dictionary containing the model's configuration, usually obtained from `model.get_config()`.
        build_if_can: bool, optional
            Whether to build the created model if input_shape was defined in the config. Defaults to False.

        Returns
        -------
        layer: Model
            The created model instance.
        """
        model_config = model_config.copy()
        model = cls()
        layers_configs = model_config['layers_configs']
        opt_config = model_config['opt_config']
        for layer_config in layers_configs:
            layer_type = layer_config['type']
            if hasattr(layers, layer_type):
                model.add(getattr(layers, layer_type).from_config(layer_config))
            elif hasattr(activations, layer_type):
                model.add(getattr(activations, layer_type).from_config(layer_config))
            else:
                raise NotImplementedError(
                    f"Creating a model that contains custom layers from a config isn't currently implemented. "
                    f"Unsupported layer type: {layer_type}.")
        if opt_config:
            opt = getattr(optimizers, opt_config['type']).from_config(opt_config)
            model.set(opt, getattr(losses, model_config['loss'])())
        input_shape = layers_configs[0].get('input_shape', None)
        if build_if_can and input_shape:
            model.build(input_shape)
        return model
